from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from flax.core.frozen_dict import freeze
from flax.traverse_util import (
    flatten_dict,
    path_aware_map,
)
from jax.tree_util import tree_leaves
from optax import (
    multi_transform,
    set_to_zero,
)

from fortuna.typing import (
    AnyKey,
    Array,
    OptaxOptimizer,
    Params,
)


def all_values_in_labels(values: Iterable, labels: Any) -> None:
    """
    Check that all values belong to the given labels.

    Parameters
    ----------
    values: Iterable
        Some iterable of values.
    labels: Any
        A collection of labels
    """
    for v in values:
        if v not in labels:
            raise ValueError(
                f"All values must belong to one of the following labels: {labels}. However, "
                f"value {v} was found."
            )


def freeze_optimizer(
    params: Params,
    optimizer: OptaxOptimizer,
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]],
) -> OptaxOptimizer:
    """
    Given an optimizer, set its gradient update to zero in correspondence to the frozen parameters specified by the
    freeze function.

    Parameters
    ----------
    params: Params
        Model parameters.
    optimizer: OptaxOptimizer
        An optimizer to freeze.
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]]
        A freeze function. It takes in input a path and its correspondent array of parameters, and it returns
        "trainable" or "frozen". If "frozen", the gradient of those parameters will be set to zero. If the freeze
        function is None, the optimizer is returned as is.

    Returns
    -------
    OptaxOptimizer
        The same optimizer but frozen where specified by the freeze function.
    """
    if freeze_fun is None:
        return optimizer
    partition_params = path_aware_map(freeze_fun, params)
    all_values_in_labels(tree_leaves(partition_params), ["trainable", "frozen"])
    partition_params = freeze(partition_params)
    partition_optimizers = {"trainable": optimizer, "frozen": set_to_zero()}
    return multi_transform(partition_optimizers, partition_params)


def get_trainable_paths(
    params: Params,
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]],
) -> Tuple[List[AnyKey], ...]:
    """
    Return a tuple of sequences of keys pointing to the trainable parameters only.

    Parameters
    ----------
    params
    freeze_fun: Callable[[Tuple[AnyKey, ...], Array], str]
        A freeze function. It takes in input a path and its correspondent array of parameters, and it returns
        "trainable" or "frozen". If "frozen", the path corresponding to those parameters will not be returned.
        If the function is `None`, all parameters are marked as trainable.

    Returns
    -------
    Tuple[List[AnyKey], ...]
        A tuple of sequences of keys pointing to the trainable parameters only.
    """
    return get_paths_with_label(
        params, freeze_fun, "trainable", allowed_labels=["frozen", "trainable"]
    )


def get_frozen_paths(
    params: Params,
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]],
) -> Tuple[List[AnyKey], ...]:
    """
    Return a tuple of sequences of keys pointing to the frozen parameters only.

    Parameters
    ----------
    params
    freeze_fun: Callable[[Tuple[AnyKey, ...], Array], str]
        A freeze function. It takes in input a path and its correspondent array of parameters, and it returns
        "trainable" or "frozen". If "frozen", the path corresponding to those parameters will not be returned.
        If the function is `None`, all parameters are marked as trainable.

    Returns
    -------
    Tuple[List[AnyKey], ...]
        A tuple of sequences of keys pointing to the frozen parameters only.
    """
    return get_paths_with_label(
        params, freeze_fun, "frozen", allowed_labels=["frozen", "trainable"]
    )


def get_paths_with_label(
    params: Union[Params, Dict],
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]],
    label: Union[str, bool],
    allowed_labels: Optional[List[Union[str, bool]]] = None,
) -> Tuple[List[AnyKey], ...]:
    paths = path_aware_map(lambda p, v: p, params)
    if freeze_fun is not None:
        conds = list(flatten_dict(path_aware_map(freeze_fun, params)).values())
        all_values_in_labels(conds, allowed_labels)
        return tuple(
            [list(p) for c, p in zip(conds, flatten_dict(paths)) if c == label]
        )
    else:
        return tuple([list(p) for p in paths])



from typing import (
    List,
    Optional,
)

from jax import (
    jit,
    lax,
    vmap,
)
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
from optax._src.base import PyTree


@jit
def kernel_stein_discrepancy_imq(
    samples: List[PyTree],
    grads: List[PyTree],
    c: float = 1.0,
    beta: float = -0.5,
) -> float:
    """Kernel Stein Discrepancy with the Inverse Multiquadric (IMQ) kernel.

    See `Gorham J. and Mackey L., 2017 <https://proceedings.mlr.press/v70/gorham17a/gorham17a.pdf>`_ for more details.

    Parameters
    ----------
        samples: List[PyTree]
            The list of `PyTree`, each representing an MCMC sample.
        grads: List[PyTree]
            The list of the corresponding density gradients.
        c: float
            :math:`c > 0` kernel bias hyperparameter.
        beta: float
            :math:`beta < 0` kernel exponent hyperparameter.

    Returns
    -------
        ksd_img: float
            The kernel Stein discrepancy value.
    """
    if not c > 0:
        raise ValueError("`c` should be > 0.")
    if not beta < 0:
        raise ValueError("`beta` should be < 0.")

    samples = ravel_pytree(samples)[0].reshape(len(samples), -1)
    grads = ravel_pytree(grads)[0].reshape(len(grads), -1)

    def _k_0(param1, param2, grad1, grad2, c, beta):
        dim = param1.shape[0]
        diff = param1 - param2
        base = c**2 + jnp.dot(diff, diff)
        kern = jnp.dot(grad1, grad2) * base**beta
        kern += -2 * beta * jnp.dot(grad1, diff) * base ** (beta - 1)
        kern += 2 * beta * jnp.dot(grad2, diff) * base ** (beta - 1)
        kern += -2 * dim * beta * (base ** (beta - 1))
        kern += -4 * beta * (beta - 1) * base ** (beta - 2) * jnp.sum(jnp.square(diff))
        return kern

    _batched_k_0 = vmap(_k_0, in_axes=(None, 0, None, 0, None, None))

    def _ksd(accum, x):
        sample1, grad1 = x
        accum += jnp.sum(_batched_k_0(sample1, samples, grad1, grads, c, beta))
        return accum, None

    ksd_sum, _ = lax.scan(_ksd, 0.0, (samples, grads))
    return jnp.sqrt(ksd_sum) / samples.shape[0]


def effective_sample_size(
    samples: List[PyTree], filter_threshold: Optional[float] = 0.0
) -> PyTree:
    """Estimate the effective sample size of a sequence.

    For a sequence of length :math:`N`, the effective sample size is defined as

    :math:`ESS(N) =  N / [ 1 + 2 * ( (N - 1) / N * R_1 + ... + 1 / N * R_{N-1} ) ]`

    where :math:`R_k` is the auto-correlation sequence,
    :math:`R_k := Cov{X_1, X_{1+k}} / Var{X_1}`

    Parameters
    ----------
        samples: List[PyTree]
            The list of `PyTree`, each representing an MCMC sample.
        filter_threshold: Optional[float]
            The cut-off value to truncate the sequence at the first index where
            the estimated auto-correlation is less than the threshold.

    Returns
    -------
        ESS: PyTree
            Parameter-wise estimates of the effective sample size.

    """
    unravel_fn = ravel_pytree(samples[0])[1]
    samples = ravel_pytree(samples)[0].reshape(len(samples), -1)

    def _autocorr(x, axis=-1, center=True):
        """Compute auto-correlation along one axis."""

        dtype = x.dtype
        shift = (-1 - axis) if axis < 0 else (len(x.shape) - 1 - axis)
        x = jnp.transpose(x, jnp.roll(jnp.arange(len(x.shape)), shift))
        if center:
            x -= x.mean(axis=-1, keepdims=True)

        # Zero pad to the next power of 2 greater than 2 * x_len
        x_len = x.shape[-1]
        pad_len = int(2.0 ** jnp.ceil(jnp.log2(x_len * 2)) - x_len)
        x = jnp.pad(x, (0, pad_len))[:-pad_len]

        # Autocorrelation is IFFT of power-spectral density
        fft = jnp.fft.fft(x.astype(jnp.complex64))
        prod = jnp.fft.ifft(fft * jnp.conj(fft))
        prod = jnp.real(prod[..., :x_len]).astype(dtype)

        # Divide to obtain an unbiased estimate of the expectation
        denominator = x_len - jnp.arange(0.0, x_len)
        res = prod / denominator
        return jnp.transpose(res, jnp.roll(jnp.arange(len(res.shape)), -shift))

    auto_cov = _autocorr(samples, axis=0)
    auto_corr = auto_cov / auto_cov[:1]

    n = len(samples)
    nk_factor = (n - jnp.arange(0.0, n)) / n
    weighted_auto_corr = nk_factor[..., None] * auto_corr

    if filter_threshold is not None:
        mask = (auto_corr < filter_threshold).astype(auto_corr.dtype)
        mask = jnp.cumsum(mask, axis=0)
        mask = jnp.maximum(1.0 - mask, 0.0)
        weighted_auto_corr *= mask

    ess = n / (-1 + 2 * weighted_auto_corr.sum(axis=0))
    return unravel_fn(ess)



from typing import (
    List,
    Optional,
    Union,
)

from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.output_calib_model.predictive.base import Predictive
from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.prob_output_layer.regression import RegressionProbOutputLayer
from fortuna.typing import Array


class RegressionPredictive(Predictive):
    def __init__(
        self,
        output_calib_manager: OutputCalibManager,
        prob_output_layer: RegressionProbOutputLayer,
    ):
        super().__init__(
            output_calib_manager=output_calib_manager,
            prob_output_layer=prob_output_layer,
        )

    def quantile(
        self,
        q: Union[float, Array, List],
        outputs: Array,
        n_samples: Optional[int] = 30,
        rng: Optional[PRNGKeyArray] = None,
        calibrated: bool = True,
    ) -> jnp.ndarray:
        """
        Estimate the quantile of the target variable given the output, with respect to the predictive distribution.

        Parameters
        ----------
        q: Union[float, Array, List]
            Quantile(s) to estimate.
        outputs : jnp.ndarray
            Model outputs.
        n_samples: Optional[int]
            Number of target samples to draw when computing quantiles.
        rng: Optional[PRNGKeyArray]
            A random number generator.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated quantiles for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.quantile(q, outputs, n_samples, rng)

    def credible_interval(
        self,
        outputs: Array,
        n_samples: int = 30,
        error: float = 0.05,
        interval_type: str = "two-tailed",
        rng: Optional[PRNGKeyArray] = None,
        calibrated: bool = True,
    ) -> jnp.ndarray:
        """
        Estimate a credible interval of the target variable given the output, with respect to the predictive
        distribution.

        Parameters
        ----------
        outputs: Array
            Model outputs.
        n_samples: int
            Number of target samples to draw for each output.
        error: float
            The interval error. This must be a number between 0 and 1, extremes included. For example,
            `error=0.05` corresponds to a 95% level of credibility.
        interval_type: str
            The interval type. We support "two-tailed" (default), "right-tailed" and "left-tailed".
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated credible interval for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.credible_interval(
            outputs, n_samples, error, interval_type, rng
        )



from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.map import MAP_NAME


class MAPPosteriorApproximator(PosteriorApproximator):
    """Maximum-A-Posteriori posterior approximator. It is responsible to define how the posterior distribution is
    approximated."""

    def __str__(self):
        return MAP_NAME



from typing import (
    Any,
    Dict,
)

from omegaconf import DictConfig


def create_input_channels(cfg: DictConfig) -> Dict[str, str]:
    base_path = cfg.dataset.base_data_path
    base_path = base_path if base_path.endswith("/") else base_path + "/"

    if (
        hasattr(cfg.dataset, "train_relative_path")
        and cfg.dataset.train_relative_path != ""
    ):
        channels = {"train": base_path + cfg.dataset.train_relative_path}
    else:
        channels = {"train": base_path}
    if (
        hasattr(cfg.dataset, "test_relative_path")
        and cfg.dataset.test_relative_path != ""
    ):
        channels.update({"test": base_path + cfg.dataset.test_relative_path})
    if (
        hasattr(cfg.dataset, "validation_relative_path")
        and cfg.dataset.validation_relative_path != ""
    ):
        channels.update(
            {"validation": base_path + cfg.dataset.validation_relative_path}
        )
    return channels


def get_base_job_name(cfg: DictConfig) -> str:
    base_job_name = (
        f"fortuna-{cfg.task.name}-{cfg.method.name}-{cfg.model.name}".replace("_", "-")
    )
    if cfg.sagemaker.job_name_suffix is not None:
        base_job_name += f"-{cfg.sagemaker.job_name_suffix}".replace("_", "-")
    return base_job_name


def get_hparams(cfg: DictConfig) -> Dict[str, Any]:
    task_hparams = {k: v for k, v in cfg.task.hparams.items()}
    model_hparams = {k: v for k, v in cfg.model.hparams.items()}
    method_hparams = {k: v for k, v in cfg.method.hparams.items()}

    hparams = dict(**task_hparams, **model_hparams, **method_hparams)

    return hparams



from fortuna.prob_model.posterior.sgmcmc.base import SGMCMCPosteriorApproximator
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld import CYCLICAL_SGLD_NAME
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_preconditioner import (
    Preconditioner,
    identity_preconditioner,
)


class CyclicalSGLDPosteriorApproximator(SGMCMCPosteriorApproximator):
    def __init__(
        self,
        n_samples: int = 10,
        n_thinning: int = 1,
        cycle_length: int = 1000,
        init_step_size: float = 1e-5,
        exploration_ratio: float = 0.25,
        preconditioner: Preconditioner = identity_preconditioner(),
    ) -> None:
        """
        Cyclical SGLD posterior approximator. It is responsible to define how the posterior distribution is approximated.

        The total number of available posterior samples depends on the number of training steps, `burnin_length` and
        `n_thinning` parameters, as well as `cycle_length` and `exploration_ratio`. In case if the number of training
        steps divides evenly by the cycle length, it can be calculated as follows:

        `n_cycles` = `n_training_steps` % `cycle_length`
        `n_sampling_steps` = (`n_cycles` * `cycle_length`) * (1 - `exploration_ratio`)
        `n_available_samples` = `n_sampling_steps` % `n_thinning`

        Setting the desired number of samples `n_samples` larger than `n_available_samples` will result in an
        exception.

        Parameters
        ----------
        n_samples: int
            The desired number of the posterior samples.
        n_thinning: int
            If `n_thinning` > 1, keep only each `n_thinning` sample during the sampling phase.
        cycle_length: int
            The length of each exploration/sampling cycle, in steps.
        init_step_size: float
            The initial step size.
        exploration_ratio: float
            The fraction of steps to allocate to the mode exploration phase.
        preconditioner: Preconditioner
            A `Preconditioner` instance that preconditions the approximator with information about the posterior distribution, if available.
        """
        super().__init__(
            n_samples=n_samples,
            n_thinning=n_thinning,
            preconditioner=preconditioner,
        )
        self.cycle_length = cycle_length
        self.init_step_size = init_step_size
        self.exploration_ratio = exploration_ratio

    def __str__(self) -> str:
        return CYCLICAL_SGLD_NAME



import copy
import importlib
import logging

from flax.core import unfreeze
from flax.traverse_util import flatten_dict
import jax
from transformers import (
    AutoConfig,
    PretrainedConfig,
)
from transformers.models.auto.auto_factory import _LazyAutoMapping
from transformers.models.auto.configuration_auto import model_type_to_module_name

logger = logging.getLogger(__name__)


def _get_model_class(config, model_mapping):
    supported_models = model_mapping[type(config)]
    if not isinstance(supported_models, (list, tuple)):
        return supported_models

    name_to_model = {model.__name__: model for model in supported_models}
    architectures = getattr(config, "architectures", [])
    for arch in architectures:
        if arch in name_to_model:
            return name_to_model[arch]
        elif f"TF{arch}" in name_to_model:
            return name_to_model[f"TF{arch}"]
        elif f"Flax{arch}" in name_to_model:
            return name_to_model[f"Flax{arch}"]

    # If not architecture is set in the config or match the supported models, the first element of the tuple is the
    # defaults.
    return supported_models[0]


class _BaseAutoSNGPModelClass:
    # Base class for auto sngp models.
    _model_mapping = None

    def __init__(self, *args, **kwargs):
        raise EnvironmentError(
            f"{self.__class__.__name__} is designed to be instantiated "
            f"using the `{self.__class__.__name__}.from_pretrained(pretrained_model_name_or_path)`"
        )

    @classmethod
    def from_config(cls, config, **kwargs):
        raise EnvironmentError(
            f"{cls.__name__} is designed to be instantiated "
            f"using the `{cls.__name__}.from_pretrained(pretrained_model_name_or_path)`."
        )

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        config = kwargs.pop("config", None)
        trust_remote_code = kwargs.pop("trust_remote_code", False)
        kwargs["_from_auto"] = True
        hub_kwargs_names = [
            "cache_dir",
            "force_download",
            "local_files_only",
            "proxies",
            "resume_download",
            "revision",
            "subfolder",
            "use_auth_token",
        ]
        hub_kwargs = {
            name: kwargs.pop(name) for name in hub_kwargs_names if name in kwargs
        }
        if not isinstance(config, PretrainedConfig):
            kwargs_copy = copy.deepcopy(kwargs)
            # ensure not to pollute the config object with torch_dtype="auto" - since it's
            # meaningless in the context of the config object - torch.dtype values are acceptable
            if kwargs_copy.get("torch_dtype", None) == "auto":
                _ = kwargs_copy.pop("torch_dtype")

            config, kwargs = AutoConfig.from_pretrained(
                pretrained_model_name_or_path,
                return_unused_kwargs=True,
                trust_remote_code=trust_remote_code,
                **hub_kwargs,
                **kwargs_copy,
            )
        if type(config) in cls._model_mapping.keys():
            model_class = _get_model_class(config, cls._model_mapping)
            model = model_class.from_pretrained(
                pretrained_model_name_or_path,
                *model_args,
                config=config,
                **hub_kwargs,
                **kwargs,
            )
            spectral_stats_params = model.params.pop("spectral_stats", None)
            batch_stats_params = model.params.pop("batch_stats", None)
            restructured_params = {"params": model.params}
            if spectral_stats_params:
                restructured_params.update({"spectral_stats": spectral_stats_params})
            if batch_stats_params:
                restructured_params.update({"batch_stats": batch_stats_params})
            # update shape of the parameters
            params_shape_tree = jax.eval_shape(
                lambda params: params, restructured_params
            )
            model._params_shape_tree = params_shape_tree
            # update required_params
            model._required_params = set(
                flatten_dict(unfreeze(params_shape_tree)).keys()
            )
            # update params
            model.params = restructured_params
            return model
        raise ValueError(
            f"Unrecognized configuration class {config.__class__} for this kind of AutoModel: {cls.__name__}.\n"
            f"At the moment, the model types that supports SNGP training are {', '.join(c.__name__ for c in cls._model_mapping.keys())}."
        )


def getattribute_from_module(module, attr):
    if attr is None:
        return None
    if isinstance(attr, tuple):
        return tuple(getattribute_from_module(module, a) for a in attr)
    if hasattr(module, attr):
        return getattr(module, attr)
    raise ValueError(f"Could not find {attr} in {module}!")


class _SNGPLazyAutoMapping(_LazyAutoMapping):
    def _load_attr_from_module(self, model_type, attr):
        module_name = model_type_to_module_name(model_type)
        if module_name not in self._modules:
            self._modules[module_name] = importlib.import_module(
                f".modeling_flax_{module_name}",
                "fortuna.prob_model.posterior.sngp.transformers.models",
            )
        return getattribute_from_module(self._modules[module_name], attr)



import abc

from fortuna.typing import Array


class NotFittedError(ValueError, AttributeError):
    """Exception class to raise if estimator is used before fitting."""


class OutOfDistributionClassifierABC:
    """
    Post-training classifier that uses the training sample embeddings coming from the model
    to score a (new) test sample w.r.t. its chance of belonging to the original training distribution
    (i.e, it is in-distribution) or not (i.e., it is out of distribution).
    """

    def __init__(self, num_classes: int):
        """
        Parameters
        ----------
        num_classes: int
            The number of classes for the in-distribution classification task.
        """
        self.num_classes = num_classes

    @abc.abstractmethod
    def fit(self, embeddings: Array, targets: Array) -> None:
        pass

    @abc.abstractmethod
    def score(self, embeddings: Array) -> Array:
        pass



from copy import deepcopy
from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Union,
)

from flax.core import FrozenDict

from fortuna.typing import AnyKey


def nested_get(
    d: Union[FrozenDict[AnyKey, Any], Dict[AnyKey, Any]], keys: List[AnyKey]
) -> Dict[AnyKey, Any]:
    """
    Get a sub-nested dictionary from a dictionary `d`. `keys` is the sequence of keys leading to the nested
    sub-dictionary.

    :param d: Dict[AnyKey, Any]
        Nested dictionary.
    :param keys: List[AnyKey
        Sequence of keys.

    :return: Dict[AnyKey, Any]
        The sub-nested dictionary.
    """
    if len(keys) == 0:
        raise ValueError("`keys` should not be empty.")
    if len(keys) == 1:
        return d[keys[0]]
    return nested_get(d[keys[0]], keys[1:])


def nested_set(
    d: Dict[AnyKey, Any],
    key_paths: Tuple[List[AnyKey], ...],
    objs: Tuple[Any],
    allow_nonexistent: bool = False,
) -> Dict[AnyKey, Any]:
    """
    Set the values of a nested dictionary for the specified sequences of keys.

    :param d: Dict
        A nested dictionary.
    :param key_paths: Tuple[List]
        Each element of the tuple is a sequence of keys defining the path to an item of the nested dictionary. The
        length of the tuple must be the same as the length of `objs`, as to each key path the corresponding object in
        `objs` will be assigned.
    :param objs: Tuple
        Each element of the tuple is an object that will be assigned to the item specified by the corresponding
        sequence of keys.
    :param allow_nonexistent: bool
        Whether to create sequence of keys that are not found in the input dictionary or throw an exception.
    """
    if type(key_paths) != tuple:
        raise TypeError("`key_paths` must be a tuple.")
    if type(objs) != tuple:
        raise TypeError("`objs` must be a tuple.")
    if len(key_paths) != len(objs):
        raise ValueError("`key_paths` and `objs` must have the same length.")

    d0 = deepcopy(d)

    for keys, o in zip(key_paths, objs):
        error_msg = f"The sequence `keys={keys}` was not found in `d`."
        d2 = d0
        for key in keys[:-1]:
            if key in d2:
                d2 = d2[key]
            elif allow_nonexistent:
                d2[key] = {}
                d2 = d2[key]
            else:
                raise KeyError(error_msg)
        if keys[-1] in d2 or allow_nonexistent:
            d2[keys[-1]] = o
        else:
            raise KeyError(error_msg)
    return d0


def nested_pair(
    d: Dict[AnyKey, Any],
    key_paths: Tuple[List[AnyKey]],
    objs: Tuple[Any],
    labels: Tuple[str, str],
) -> Dict[AnyKey, Any]:
    """
    Replace the values of a nested dictionary at the specified sequences of keys `keys` with dictionaries including both
    the values in the original dictionary and the new objects in `objs`. The keys of these dictionaries are given by
    `labels`.

    :param d: Dict
        A nested dictionary.
    :param key_paths: Tuple[List]
        Each element of the tuple is a sequence of keys defining the path to an item of the nested dictionary. The
        length of the tuple must be the same as the length of `objs`, as to each key path the corresponding object in
        `objs` will be assigned.
    :param objs: Tuple
        Each element of the tuple is an object that will be included in the item specified by the corresponding
        sequence of keys.
    :param labels: Tuple[str, str]
        Labels for the values of `d` and of `objs`, respectively.
    """
    if type(key_paths) != tuple:
        raise TypeError("`key_paths` must be a tuple.")
    if type(objs) != tuple:
        raise TypeError("`values` must be a tuple.")
    if len(key_paths) != len(objs):
        raise ValueError("`key_paths` and `values` must have the same length.")
    if len(labels) != 2:
        raise ValueError("The length of `labels` must be exactly 2.")

    d0 = deepcopy(d)

    for keys, o in zip(key_paths, objs):
        error_msg = f"The sequence `keys={keys}` was not found in `d`."
        d2 = d0
        for key in keys[:-1]:
            if key in d2:
                d2 = d2[key]
            else:
                raise KeyError(error_msg)
        if keys[-1] in d2:
            d2[keys[-1]] = {labels[0]: d2[keys[-1]], labels[1]: o}
        else:
            raise KeyError(error_msg)
    return d0


def nested_unpair(
    d: Dict[AnyKey, Any], key_paths: Tuple[List[AnyKey], ...], labels: Tuple[str, str]
) -> Tuple[Dict[AnyKey, Any], Dict[AnyKey, Any]]:
    """
    Form two dictionaries out of the initial dictionary `d`. In correspondence to the sequences of keys in `key_paths`,
     the first and second dictionary will get the values marked by `labels`. Otherwise, the first dictionary takes the
     values of `d`, while the second dictionary does not take any.

    :param d: Dict[AnyKey, Any]
        A nested dictionary.
    :param key_paths: Tuple[List[AnyKey]]
        Each element of the tuple is a sequence of keys defining the path to an item of the nested dictionary. In
        correspondence to each key path, the value must be a dictionary with `labels` as keys.
    :param labels: Tuple[str, str]
        Labels indicating the keys of the objects going into the first and second dictionaries, respectively.

    :return Tuple[Dict[AnyKey, Any], Dict[AnyKey, Any]]
        The two dictionaries.
    """
    if type(key_paths) != tuple:
        raise TypeError("`key_paths` must be a tuple.")
    if len(labels) != 2:
        raise ValueError("The length of `labels` must be exactly 2.")

    d01 = deepcopy(d)
    d02 = dict()

    for keys in key_paths:
        error_msg = f"The sequence `keys={keys}` was not found in `d`."
        d21 = d01
        d22 = d02
        for key in keys[:-1]:
            if key in d21:
                d21 = d21[key]
                if key not in d22:
                    d22[key] = dict()
                d22 = d22[key]
            else:
                raise KeyError(error_msg)
        if keys[-1] in d21:
            d22[keys[-1]] = d21[keys[-1]][labels[1]]
            d21[keys[-1]] = d21[keys[-1]][labels[0]]
        else:
            raise KeyError(error_msg)

    return d01, d02


def find_one_path_to_key(
    d: Dict[AnyKey, Any], key: AnyKey, keys: Tuple[AnyKey] = ()
) -> Tuple[AnyKey]:
    if hasattr(d, "items"):
        for k, v in d.items():
            if k == key:
                keys += (k,)
                return keys
            if isinstance(v, (FrozenDict, dict)):
                tmp = find_one_path_to_key(v, key, keys + (k,))
                if key in tmp:
                    return tmp
    return keys


def nested_update(
    d: Dict[AnyKey, Any], *updating_d: Dict[AnyKey, Any]
) -> Dict[AnyKey, Any]:
    """
    Update a nested dictionary `d` with the content of an other dictionary `updating_d`.
    """
    # from https://github.com/pydantic/pydantic/blob/9d631a3429a66f30742c1a52c94ac18ec6ba848d/pydantic/utils.py#L198
    updated_mapping = d.copy()
    for updating_mapping in updating_d:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = nested_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping



from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
)

from flax.core import FrozenDict
from flax.training import dynamic_scale
import jax.numpy as jnp

from fortuna.training.train_state import TrainState
from fortuna.typing import (
    CalibMutable,
    CalibParams,
    Mutable,
    OptaxOptimizer,
    Params,
)
from fortuna.utils.strings import convert_string_to_tuple


class PosteriorState(TrainState):
    """
    A posterior distribution state. This includes all the parameters and mutable objects that characterize an
    approximation of the posterior distribution.
    """

    params: Params
    mutable: Optional[Mutable] = None
    calib_params: Optional[CalibParams] = None
    calib_mutable: Optional[CalibMutable] = None
    grad_accumulated: Optional[jnp.ndarray] = None
    dynamic_scale: Optional[dynamic_scale.DynamicScale] = None
    encoded_name: tuple = convert_string_to_tuple("PosteriorState")

    @classmethod
    def init(
        cls,
        params: Params,
        mutable: Optional[Mutable] = None,
        optimizer: Optional[OptaxOptimizer] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        grad_accumulated: Optional[jnp.ndarray] = None,
        dynamic_scale: Optional[dynamic_scale.DynamicScale] = None,
        **kwargs,
    ) -> Any:
        """
        Initialize a posterior distribution state.

        Parameters
        ----------
        params : Params
            The parameters characterizing an approximation of the posterior distribution.
        optimizer : Optional[OptaxOptimizer]
            An Optax optimizer associated with the posterior state.
        mutable : Optional[Mutable]
            The mutable objects characterizing an approximation of the posterior distribution.
        calib_params : Optional[CalibParams]
            The parameters objects characterizing an approximation of the posterior distribution.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects characterizing an approximation of the posterior distribution.
        grad_accumulated : Optional[jnp.ndarray]
            The gradients accumulated in consecutive training steps (used only when `gradient_accumulation_steps > 1`).
        dynamic_scale: Optional[dynamic_scale.DynamicScale]
            Dynamic loss scaling for mixed precision gradients.
        Returns
        -------
        Any
            A posterior distribution state.
        """
        return cls(
            apply_fn=None,
            params=params,
            opt_state=kwargs["opt_state"]
            if optimizer is None and "opt_state" in kwargs
            else optimizer.init(params),
            mutable=mutable,
            step=kwargs.get("step", 0),
            tx=optimizer,
            calib_params=calib_params,
            calib_mutable=calib_mutable,
            dynamic_scale=dynamic_scale,
            grad_accumulated=grad_accumulated,
            **{
                k: v
                for k, v in kwargs.items()
                if k not in ["opt_state", "apply_fn", "tx", "step"]
            },
        )

    @classmethod
    def init_from_dict(
        cls,
        d: Dict,
        optimizer: Optional[OptaxOptimizer] = None,
        **kwargs,
    ) -> PosteriorState:
        """
        Initialize a posterior distribution state from a dictionary.

        Parameters
        ----------
        d : Dict
            A dictionary including attributes of the posterior state.
        optimizer : Optional[OptaxOptimizer]
            An optax optimizer to assign to the posterior state.

        Returns
        -------
        PosteriorState
            A posterior state.
        """
        kwargs = {
            **kwargs,
            **{
                k: v
                for k, v in d.items()
                if k
                not in [
                    "params",
                    "mutable",
                    "optimizer",
                    "calib_params",
                    "calib_mutable",
                ]
            },
        }
        return cls.init(
            FrozenDict(d["params"]),
            FrozenDict(d["mutable"]) if d["mutable"] is not None else None,
            optimizer,
            FrozenDict(d.get("calib_params"))
            if d["calib_params"] is not None
            else None,
            FrozenDict(d.get("calib_mutable"))
            if d["calib_mutable"] is not None
            else None,
            **kwargs,
        )



from __future__ import annotations

from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import jax.numpy as jnp

from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import (
    Array,
    OptaxOptimizer,
)
from fortuna.utils.strings import convert_string_to_tuple


class SWAGState(PosteriorState):
    """
    Attributes
    ----------
    encoded_name: jnp.ndarray
        SWAG state name encoded as an array.
    mean: Optional[jnp.ndarray]
        Mean of the posterior approximation.
    std: Optional[jnp.ndarray]
        Diagonal standard deviation of the posterior approximation.
    dev: Optional[jnp.ndarray]
        Deviation term of the covariance matrix of the posterior approximation.
    """

    mean: Optional[jnp.ndarray] = None
    std: Optional[jnp.ndarray] = None
    dev: Optional[jnp.ndarray] = None
    encoded_name: tuple = convert_string_to_tuple("SWAGState")
    _encoded_which_params: Optional[Dict[str, List[Array]]] = None

    @classmethod
    def convert_from_map_state(
        cls, map_state: MAPState, optimizer: OptaxOptimizer
    ) -> SWAGState:
        """
        Convert a MAP state into a SWAG state.

        Parameters
        ----------
        map_state: MAPState
            A MAP posterior state.
        optimizer: OptaxOptimizer
            An Optax optimizer.

        Returns
        -------
        SWAGState
            A SWAG state.
        """
        return SWAGState.init(
            params=map_state.params,
            mutable=map_state.mutable,
            optimizer=optimizer,
            calib_params=map_state.calib_params,
            calib_mutable=map_state.calib_mutable,
        )

    def update(self, variables: Dict[str, Any]) -> SWAGState:
        """
        Update the SWAG state.

        Parameters
        ----------
        variables: Dict[str, Any]
            The attributes to update and their values.

        Returns
        -------
        SWAGState
            Updated SWAG state.
        """
        unchanged_keys = {k: v for k, v in vars(self).items() if k not in variables}
        return self.replace(**unchanged_keys, **variables)



from typing import Optional

import flax.linen as nn
from flax.linen.initializers import Initializer
import jax.numpy as jnp

from fortuna.typing import Array


class ScalarConstantModel(nn.Module):
    r"""
    A scalar constant model, that is :math:`f(\theta, x) = \theta`, with :math:`\theta\in\mathbb{R}`. The scalar value
    will be broadcasted to the output dimension.

    Parameters
    ----------
    output_dim: int
        The output model dimension.
    initializer_fun: Optional[Initializer]
        Function to initialize the model parameters.
        This must be one of the available options in :code:`flax.linen.initializers`.
    """
    output_dim: int
    initializer_fun: Optional[Initializer] = nn.initializers.zeros

    @nn.compact
    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        scalar = self.param("scalar", self.initializer_fun, (1,))
        return jnp.broadcast_to(scalar, shape=(x.shape[0], self.output_dim))



from fortuna.calibration.binary_classification.temp_scaling.bias_binary_temp_scaling import (
    BiasBinaryClassificationTemperatureScaling,
)
from fortuna.calibration.binary_classification.temp_scaling.brier_binary_temp_scaling import (
    BrierBinaryClassificationTemperatureScaling,
)
from fortuna.calibration.binary_classification.temp_scaling.crossentropy_binary_temp_scaling import (
    CrossEntropyBinaryClassificationTemperatureScaling,
)
from fortuna.calibration.binary_classification.temp_scaling.f1_temp_scaling import (
    F1BinaryClassificationTemperatureScaling,
)
from fortuna.calibration.classification.temp_scaling.base import (
    ClassificationTemperatureScaling,
)



from typing import Any

import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array


class LeNet5(nn.Module):
    """
    A LeNet-5 network [LeCun et al., 1989](http://yann.lecun.com/exdb/publis/pdf/lecun-89e.pdf). Please refer to
    :class:`~fortuna.prob_model.model.base.Model` for the internal methods.

    Attributes
    ----------
    output_dim: int
        The output model dimension.
    dtype: Any
        Layers' dtype.
    """

    output_dim: int
    dtype: Any = jnp.float32

    def setup(self):
        self.dfe_subnet = LeNet5DeepFeatureExtractorSubNet(dtype=self.dtype)
        self.output_subnet = LeNet5OutputSubNet(
            output_dim=self.output_dim, dtype=self.dtype
        )

    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: Array
            Inputs.

        jnp.ndarray
            Model outputs.
        """
        x = self.dfe_subnet(x)
        x = self.output_subnet(x)
        return x


class LeNet5DeepFeatureExtractorSubNet(nn.Module):
    """
    Deep feature extractor sub-network of a LeNet-5.

    Attributes
    ----------
    dtype: Any
        Layers' dtype.
    """

    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, x: Array):
        """
        Forward pass.

        Parameters
        ----------
        x: Array
            Inputs.

        jnp.ndarray
            Output of the hidden layers.
        """
        x = nn.Conv(
            features=6,
            kernel_size=(5, 5),
            strides=(1, 1),
            padding="valid",
            dtype=self.dtype,
        )(x)
        x = nn.relu(x)
        x = nn.max_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = nn.Conv(
            features=16,
            kernel_size=(5, 5),
            strides=(1, 1),
            padding="valid",
            dtype=self.dtype,
        )(x)
        x = nn.relu(x)
        x = nn.max_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = x.reshape((x.shape[0], -1))  # flatten
        x = nn.Dense(features=120, dtype=self.dtype)(x)
        x = nn.relu(x)
        x = nn.Dense(features=84, dtype=self.dtype)(x)
        return x


class LeNet5OutputSubNet(nn.Module):
    """
    Output sub-network of a LeNet-5.

    Attributes
    ----------
    output_dim: int
        The output model dimension.
    dtype: Any
        Layers' dtype.
    """

    output_dim: int
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, x: jnp.ndarray):
        """
        Forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Outputs of the deep feature extractor sub-network.

        Returns
        -------
        jnp.ndarray
            Model outputs.
        """
        x = nn.relu(x)
        x = nn.Dense(features=self.output_dim, dtype=self.dtype)(x)
        return x



import jax.numpy as jnp

from fortuna.conformal.regression.base import ConformalRegressor
from fortuna.typing import Array


class JackknifePlusConformalRegressor(ConformalRegressor):
    """
    This class implements the jackknife+ method introduced in
    `Barber et al., 2021 <https://www.stat.cmu.edu/~ryantibs/papers/jackknife.pdf>`__. Given a desired coverage of
    :math:`1-α`, jackknife+ guarantees a minimal coverage of :math:`1 - 2α`.
    """

    def conformal_interval(
        self,
        loo_val_outputs: Array,
        loo_val_targets: Array,
        loo_test_outputs: Array,
        error: float,
    ) -> jnp.ndarray:
        """
        Coverage interval of each of the test inputs, at the desired coverage error. This is supported only for
        one-dimensional target variables.

        Parameters
        ----------
        loo_val_outputs: Array
            Outputs of the models used during leave-out-out evaluated at their respective left-out validation inputs.
            More precisely, the i-th element of :code: `loo_val_outputs` is a model that has been trained upon all the
            training data but the i-th data point, and evaluated at the i-th input, for all training data points.
        loo_val_targets: Array
            The array of target variables of the left-out data points in the leave-one-out procedure described for
            `loo_val_outputs`.
        loo_test_outputs: Array
            Outputs of the models used during leave-one-out evaluated at the test inputs. More precisely, consider
            the same leave-one-out procedure as the one used for :code:`loo_val_outputs`. Then the i-th element of
            the first dimension of :code:`loo_test_outputs` represents the outputs of the model that has been trained
            upon all the training data but the i-th data point, and evaluated at the test inputs. The second dimension
            of :code:`loo_test_outputs` is over the different test inputs.
        error: float
            The desired coverage error. This must be a scalar between 0 and 1, extremes included.
        Returns
        -------
        jnp.ndarray
            The conformal intervals. The two components of the second axis correspond to the left and right interval
            bounds.
        """
        if loo_val_outputs.shape[0] != loo_val_targets.shape[0]:
            raise ValueError(
                "The first dimension of `loo_val_outputs` and `loo_val_targets` must coincide. However, "
                f"{loo_val_outputs.shape[0]} and {loo_val_targets.shape[0]} were found, respectively."
            )
        if loo_val_outputs.ndim == 1:
            loo_val_outputs = loo_val_outputs[:, None]
        elif loo_val_outputs.shape[1] != 1:
            raise ValueError(
                "This method is supported only for scalar model outputs only. However, `loo_val_outputs` has second "
                "dimension greater than 1."
            )
        if loo_val_targets.ndim == 1:
            loo_val_targets = loo_val_targets[:, None]
        elif loo_val_targets.shape[1] != 1:
            raise ValueError(
                "This method is supported only for scalar target variables. However, `loo_val_targets` "
                "has second dimension greater than 1."
            )
        if loo_test_outputs.ndim < 2:
            raise ValueError("`loo_test_outputs` must have at least two dimensions.")
        elif loo_test_outputs.ndim == 2:
            loo_test_outputs = loo_test_outputs[:, :, None]
        elif loo_test_outputs.shape[2] != 1:
            raise ValueError(
                "This method is supported only for scalar model outputs only. However, `loo_test_outputs` "
                "has last dimension greater than 1."
            )

        r = jnp.abs(loo_val_targets - loo_val_outputs)
        left = loo_test_outputs - r[:, None]
        right = loo_test_outputs + r[:, None]

        qleft = jnp.quantile(left, q=error, axis=0)
        qright = jnp.quantile(right, q=1 - error, axis=0)
        return jnp.array(list(zip(qleft, qright))).squeeze(2)



from fortuna.model.constant import ConstantModel
from fortuna.model.hyper import HyperparameterModel
from fortuna.model.lenet import LeNet5
from fortuna.model.linear import LinearModel
from fortuna.model.mlp import MLP
from fortuna.model.resnet import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    ResNet200,
)
from fortuna.model.scalar_constant import ScalarConstantModel
from fortuna.model.scalar_hyper import ScalarHyperparameterModel
from fortuna.model.wideresnet import WideResNet28_10



from collections import OrderedDict

from transformers.models.auto.configuration_auto import CONFIG_MAPPING_NAMES

from fortuna.prob_model.posterior.sngp.transformers.auto_factory import (
    _BaseAutoSNGPModelClass,
    _SNGPLazyAutoMapping,
)

FLAX_SNGP_MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING_NAMES = OrderedDict(
    [
        # Model for Sequence Classification with SNGP mapping
        ("bert", "FlaxSNGPBertExtractorForSequenceClassification"),
        ("distilbert", "FlaxSNGPDistilBertExtractorForSequenceClassification"),
        ("roberta", "FlaxSNGPRobertaExtractorForSequenceClassification"),
    ]
)

FLAX_SNGP_MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING = _SNGPLazyAutoMapping(
    CONFIG_MAPPING_NAMES, FLAX_SNGP_MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING_NAMES
)


class FlaxAutoSNGPModelForSequenceClassification(_BaseAutoSNGPModelClass):
    _model_mapping = FLAX_SNGP_MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING



from __future__ import annotations

from typing import (
    Dict,
    Optional,
    Union,
)

from flax.core import FrozenDict

from fortuna.typing import (
    CalibMutable,
    CalibParams,
)


class OutputCalibManagerState:
    params: CalibParams
    mutable: Optional[CalibMutable] = None

    def __init__(self, params: CalibParams, mutable: Optional[CalibMutable] = None):
        """
        An model manager state class.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        """
        self.params = params
        self.mutable = mutable

    @classmethod
    def init_from_dict(cls, d: Union[Dict, FrozenDict]) -> OutputCalibManagerState:
        """
        Initialize an output calibration manager state from a dictionary.

        Parameters
        ----------
        d : Union[Dict, FrozenDict]
            A dictionary with as keys the calibrators and as values their initializations.

        Returns
        -------
        OutputCalibManagerState
            An output calibration manager state.
        """
        params = FrozenDict(
            {
                k: FrozenDict({"params": v["params"] if v else None})
                for k, v in d.items()
            }
        )
        mutable = {calib_name: {} for calib_name in d}
        for name, variables in d.items():
            if variables:
                for var_name, var_obj in variables.items():
                    if var_name != "params":
                        mutable[name][var_name] = var_obj
        mutable = FrozenDict(
            {name: v if len(v) > 0 else None for name, v in mutable.items()}
        )
        return cls(params=params, mutable=mutable)



import os
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from fortuna.prob_model.posterior.posterior_state_repository import (
    PosteriorStateRepository,
)
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import (
    OptaxOptimizer,
    Path,
)


class PosteriorMultiStateRepository:
    def __init__(self, size: int, checkpoint_dir: Optional[Path] = None):
        self.size = size
        self.state = [
            PosteriorStateRepository(
                checkpoint_dir=os.path.join(checkpoint_dir, str(i))
                if checkpoint_dir
                else None
            )
            for i in range(size)
        ]

    def get(
        self,
        i: int = None,
        checkpoint_path: Optional[Path] = None,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Union[List[PosteriorState], PosteriorState]:
        def _get(_i):
            return self.state[_i].get(
                checkpoint_path=checkpoint_path,
                optimizer=optimizer,
                prefix=prefix,
                **kwargs,
            )

        if i is not None:
            return _get(i)
        state = []
        for i in range(self.size):
            state.append(_get(i))
        return state

    def put(
        self,
        state: PosteriorState,
        i: int = None,
        checkpoint_path: Optional[Path] = None,
        keep: int = 1,
        prefix: str = "checkpoint_",
    ) -> None:
        def _put(_i):
            return self.state[_i].put(
                state=state, checkpoint_path=checkpoint_path, keep=keep, prefix=prefix
            )

        if i is not None:
            _put(i)
        else:
            for i in range(self.size):
                state.append(_put(i))

    def pull(
        self,
        i: int = None,
        checkpoint_path: Path = None,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> PosteriorState:
        def _pull(_i):
            return self.state[_i].pull(
                checkpoint_path=checkpoint_path,
                optimizer=optimizer,
                prefix=prefix,
                **kwargs,
            )

        if i is not None:
            return _pull(i)
        state = []
        for i in range(self.size):
            state.append(_pull(i))
        return state

    def update(
        self,
        variables: Dict,
        i: int = None,
        checkpoint_path: Path = None,
        optimizer: Optional[OptaxOptimizer] = None,
        keep: int = 1,
        prefix: str = "checkpoint_",
        **kwargs,
    ):
        def _update(_i):
            self.state[_i].update(
                variables=variables,
                checkpoint_path=checkpoint_path,
                optimizer=optimizer,
                keep=keep,
                prefix=prefix,
                **kwargs,
            )

        if i is not None:
            _update(i)
        else:
            for i in range(self.size):
                _update(i)

    def extract(
        self,
        keys: List[str],
        i: int = None,
        checkpoint_path: Optional[Path] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Union[Dict, List[Dict]]:
        def _extract(_i):
            return self.state[_i].extract(
                keys=keys, checkpoint_path=checkpoint_path, prefix=prefix, **kwargs
            )

        if i is not None:
            return _extract(i)
        dicts = []
        for i in range(self.size):
            dicts.append(_extract(i))
        return dicts

    def extract_calib_keys(
        self,
        checkpoint_path: Optional[Path] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Dict:
        return self.extract(
            ["calib_params", "calib_mutable"], 0, checkpoint_path, prefix, **kwargs
        )



from fortuna.output_calib_model.output_calib_mixin import (
    WithOutputCalibCheckpointingMixin,
)
from fortuna.training.train_state_repository import TrainStateRepository


class OutputCalibStateRepository(
    WithOutputCalibCheckpointingMixin, TrainStateRepository
):
    pass



from typing import Optional

import jax.numpy as jnp

from fortuna.typing import Array


class BatchMVPMixin:
    def pinball_loss(self, values: Array, scores: Array, coverage: float) -> Array:
        """
        The pinball loss between the model evaluations and the scores.

        Parameters
        ----------
        values: Array
            The model evaluations.
        scores: Array
            The scores.
        coverage: float
            The target coverage.

        Returns
        -------
        Array
            The pinball loss evaluation.
        """
        return self._loss_fn(values, scores, coverage=coverage)

    def _loss_fn(
        self, values: Array, scores: Array, coverage: Optional[float] = None
    ) -> Array:
        if scores.ndim == 2 and values.ndim == 1:
            scores = scores[:, 0]
        if coverage is None:
            coverage = self._coverage
        diff = scores - values
        return jnp.mean(
            diff * coverage * (scores > values)
            - diff * (1 - coverage) * (scores <= values)
        )



from typing import (
    List,
    Union,
)

import numpy as np


class GroupingModel:
    """
    Grouping model based on clustering of embeddings.
    """

    def fit(
        self,
        embeddings: np.ndarray,
        clustering_models: List,
    ) -> None:
        """
        Fit the model.

        Parameters
        ----------
        embeddings: np.ndarray
            Embeddings of inputs.
        clustering_models: List
            A list of clustering models. Each clustering model must include the following method:

            - `fit`, to fit the model;
            - `bic`, to compute the BIC;
            - `predict_proba`, to get the predicted probabilities;
            - `predict`, to get the predictions.
            An example of valid clustering model is `sklearn.mixture.GaussianMixture`.
        """
        self._store_embeddings_stats(embeddings)
        embeddings = self._normalize(embeddings)

        self._fit_best_model(embeddings, models=clustering_models)

    def predict_proba(self, embeddings: np.ndarray) -> np.ndarray:
        """
        For each input, predict the probability of belonging to each cluster.

        Parameters
        ----------
        embeddings: np.ndarray
            Embeddings of inputs.

        Returns
        -------
        np.ndarray
            Predicted probabilities.
        """
        if self._clustering_model is None:
            raise ValueError("The `fit` method must be run first.")
        embeddings = self._normalize(embeddings)
        return self._clustering_model.predict_proba(embeddings)

    def soft_predict(
        self, embeddings: np.ndarray, quantile_proba_threshold: float = 0.8
    ) -> np.ndarray:
        """
        For each input, predict which clusters the inputs are most likely to belong to.

        Parameters
        ----------
        embeddings: np.ndarray
            Embeddings of inputs.
        quantile_proba_threshold: float
            The threshold at which to compute quantiles of the group scores.

        Returns
        -------
        np.ndarray
            An array of bools determining whether an input is predicted to belong to a cluster or not.
        """
        probs = self.predict_proba(embeddings=embeddings)
        quantiles = self.compute_quantile(probs, quantile_proba_threshold)
        return probs > quantiles[None]

    def hard_predict(
        self,
        embeddings: np.ndarray,
    ) -> np.ndarray:
        """
        For each input, predict the most likely cluster it belongs to.

        Parameters
        ----------
        embeddings: np.ndarray
            Embeddings of inputs.

        Returns
        -------
        np.ndarray
            An array of bools determining whether an input is predicted to belong to a cluster or not.
            Exactly one True will be given for each input.
        """
        probs = self.predict_proba(embeddings=embeddings)

        bool_preds = np.zeros_like(probs, dtype=bool)
        bool_preds[np.arange(len(probs)), np.argmax(probs, axis=1)] = True
        return bool_preds

    @property
    def clustering_model(self):
        return self._clustering_model

    @staticmethod
    def compute_quantile(probs, threshold: float) -> Union[float, np.ndarray]:
        return np.quantile(probs, threshold, axis=0)

    def _normalize(self, embeddings: np.ndarray) -> np.ndarray:
        if self._mean is None or self._std is None:
            raise ValueError("The `fit` method must be run first.")
        embeddings = np.copy(embeddings)
        embeddings -= self._mean
        embeddings /= self._std
        return embeddings

    def _store_embeddings_stats(self, embeddings):
        self._mean = np.mean(embeddings, axis=0, keepdims=True)
        self._std = np.std(embeddings, axis=0, keepdims=True)

    def _fit_best_model(self, embeddings: np.ndarray, models: List) -> None:
        best_bic = np.inf
        for model in models:
            model = model.fit(embeddings)
            bic = model.bic(embeddings)
            if bic < best_bic:
                best_bic = bic
                best_model = model
        self._clustering_model = best_model



import importlib
import logging
from typing import (
    Dict,
    Optional,
    Type,
)

import flax.linen as nn

from fortuna.data.loader import DataLoader
from fortuna.likelihood.classification import ClassificationLikelihood
from fortuna.model.model_manager.classification import (
    ClassificationModelManager,
    SNGPClassificationModelManager,
)
from fortuna.model.model_manager.name_to_model_manager import (
    ClassificationModelManagers,
)
from fortuna.model_editor.base import ModelEditor
from fortuna.output_calibrator.classification import ClassificationTemperatureScaler
from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.prob_model.base import ProbModel
from fortuna.prob_model.calib_config.base import CalibConfig
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.posterior_approximations import (
    PosteriorApproximations,
)
from fortuna.prob_model.posterior.swag.swag_approximator import (
    SWAGPosteriorApproximator,
)
from fortuna.prob_model.predictive.classification import ClassificationPredictive
from fortuna.prob_model.prior import IsotropicGaussianPrior
from fortuna.prob_model.prior.base import Prior
from fortuna.prob_output_layer.classification import (
    ClassificationMaskedProbOutputLayer,
    ClassificationProbOutputLayer,
)
from fortuna.typing import Status
from fortuna.utils.data import (
    get_input_shape,
    get_inputs_from_shape,
)


class ProbClassifier(ProbModel):
    def __init__(
        self,
        model: nn.Module,
        prior: Prior = IsotropicGaussianPrior(),
        posterior_approximator: PosteriorApproximator = SWAGPosteriorApproximator(),
        output_calibrator: Optional[nn.Module] = ClassificationTemperatureScaler(),
        model_editor: Optional[ModelEditor] = None,
        seed: int = 0,
    ):
        r"""
        A probabilistic classifier class.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. The outputs must correspond to
            the logits of a softmax probability vector. The output dimension must be the same as the number of classes.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`f(w, x)`, where each component of :math:`f` corresponds to one of the classes.
        prior : Prior
            A prior distribution object. The default is an isotropic standard Gaussian. Let :math:`w` be the random
            model parameters. Then the prior is defined by a distribution :math:`p(w)`.
        posterior_approximator : PosteriorApproximator
            A posterior approximation method. The default method is SWAG.
        output_calibrator : Optional[nn.Module]
            An output calibrator object. The default is temperature scaling for classification, which rescales the
            logits with a scalar temperature parameter. Given outputs :math:`o` of the model manager, the output
            calibrator is described by a function :math:`g(\phi, o)`, where `phi` are deterministic
            calibration parameters.
        model_editor : ModelEditor
            A model_editor objects. It takes the forward pass and transforms the outputs.
        seed: int
            A random seed.

        Attributes
        ----------
        model : nn.Module
            See `model` in `Parameters`.
        model_manager : ClassificationModelManager
            This object orchestrates the model's forward pass.
        output_calibrator : nn.Module
            See `output_calibrator` in `Parameters`.
        prob_output_layer : ClassificationProbOutputLayer
            This object characterizes the distribution of target variable given the calibrated outputs. It is defined
            by :math:`p(y|o)=\text{Categorical}(y|p=softmax(o))`,
            where :math:`o` denotes the calibrated outputs and :math:`y` denotes a target variable.
        likelihood : ClassificationLikelihood
            The likelihood function. This is defined by
            :math:`p(y|w, \phi, x) = \text{Categorical}(y|p=\text{softmax}(g(\phi, f(w, x)))`.
        prior : Prior
            See `prior` in `Parameters`.
        joint : Joint
            This object describes the joint distribution of the target variables and the random parameters
            given the input variables and the calibration parameters, that is :math:`p(y, w|x, \phi)`.
        posterior_approximator : PosteriorApproximator
            See `posterior_approximator` in `Parameters`.
        posterior : Posterior
            This is the posterior approximation of the random parameters given the training data and the
            calibration parameters, that is :math:`p(w|\mathcal{D}, \phi)`, where :math:`\mathcal{D}` denotes the
            training data set and :math:`\phi` the calibration parameters.
        predictive : ClassificationPredictive
            This denotes the predictive distribution, that is :math:`p(y|\phi, x, \mathcal{D})`. Its statistics are
            approximated via a Monte Carlo approach by sampling from the posterior approximation.
        """
        self.model = model
        self.prior = prior
        self.output_calibrator = output_calibrator

        self.output_calib_manager = OutputCalibManager(
            output_calibrator=output_calibrator
        )
        self.prob_output_layer = self._get_prob_output_layer(model)

        model_manager_cls = getattr(
            ClassificationModelManagers, posterior_approximator.__str__()
        ).value
        self.model_manager = self._get_model_manager(
            model, model_editor, model_manager_cls, posterior_approximator
        )

        self.likelihood = ClassificationLikelihood(
            self.model_manager, self.prob_output_layer, self.output_calib_manager
        )
        self.joint = Joint(self.prior, self.likelihood)

        self.posterior = getattr(
            PosteriorApproximations, posterior_approximator.__str__()
        ).value(joint=self.joint, posterior_approximator=posterior_approximator)
        self.predictive = ClassificationPredictive(self.posterior)

        super().__init__(seed=seed)

    def _get_prob_output_layer(self, model: nn.Module) -> ClassificationProbOutputLayer:
        try:
            # import modules if available
            transformers_flax_auto_module = importlib.import_module(
                "transformers.models.auto.modeling_flax_auto"
            )
            FLAX_MODEL_FOR_MASKED_LM_MAPPING_NAMES = list(
                getattr(
                    transformers_flax_auto_module,
                    "FLAX_MODEL_FOR_MASKED_LM_MAPPING_NAMES",
                ).values()
            )
            if str(model.__class__.__name__) in FLAX_MODEL_FOR_MASKED_LM_MAPPING_NAMES:
                prob_output_layer = ClassificationMaskedProbOutputLayer()
            else:
                prob_output_layer = ClassificationProbOutputLayer()
        except ModuleNotFoundError:
            prob_output_layer = ClassificationProbOutputLayer()
        return prob_output_layer

    def _get_model_manager(
        self,
        model: nn.Module,
        model_editor: ModelEditor,
        model_manager_cls: Type,
        posterior_approximator: PosteriorApproximator,
    ) -> ClassificationModelManager:
        try:
            # import modules if available
            transformers_module = importlib.import_module("transformers")
            fortuna_transformers_classification_module = importlib.import_module(
                "fortuna.model.model_manager.transformers.classification"
            )
            # import relevant classes
            FlaxPreTrainedModel = getattr(transformers_module, "FlaxPreTrainedModel")
            SNGPHuggingFaceClassificationModelManager = getattr(
                fortuna_transformers_classification_module,
                "SNGPHuggingFaceClassificationModelManager",
            )
            HuggingFaceClassificationModelManager = getattr(
                fortuna_transformers_classification_module,
                "HuggingFaceClassificationModelManager",
            )
            # load model manager
            if (
                isinstance(model, FlaxPreTrainedModel)
                and model_manager_cls == SNGPClassificationModelManager
            ):
                model_manager = SNGPHuggingFaceClassificationModelManager(
                    model=model,
                    model_editor=model_editor,
                    **posterior_approximator.posterior_method_kwargs,
                )
            elif isinstance(model, FlaxPreTrainedModel):
                model_manager = HuggingFaceClassificationModelManager(
                    model, model_editor=model_editor
                )
            else:
                model_manager = model_manager_cls(
                    model=model,
                    model_editor=model_editor,
                    **posterior_approximator.posterior_method_kwargs,
                )
        except ModuleNotFoundError as e:
            logging.warning(
                "No module named 'transformer' is installed. "
                "If you are not working with models from the `transformers` library ignore this warning, otherwise "
                "install the optional 'transformers' dependency of Fortuna using poetry. You can do so by entering: "
                "`poetry install --extras 'transformers'`."
            )
            model_manager = model_manager_cls(
                model=model,
                model_editor=model_editor,
                **posterior_approximator.posterior_method_kwargs,
            )
        return model_manager

    def _check_output_dim(self, data_loader: DataLoader):
        if data_loader.size == 0:
            raise ValueError(
                """`data_loader` is either empty or incorrectly constructed."""
            )
        output_dim = data_loader.num_unique_labels
        for x, y in data_loader:
            input_shape = get_input_shape(x)
            break
        s = self.joint.init(input_shape)
        inputs = get_inputs_from_shape(input_shape)
        outputs = self.model_manager.apply(
            params=s.params, inputs=inputs, mutable=s.mutable
        )
        model_output_dim = (
            outputs[0].shape[-1]
            if isinstance(outputs, (list, tuple))
            else outputs.shape[-1]
        )
        if model_output_dim != output_dim:
            raise ValueError(
                f"""The outputs dimension of `model` must correspond to the number of different classes
            in the target variables of `_data_loader`. However, {model_output_dim} and {output_dim} were found,
            respectively."""
            )

    def train(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        calib_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        calib_config: CalibConfig = CalibConfig(),
        **fit_kwargs,
    ) -> Dict[str, Status]:
        self._check_output_dim(train_data_loader)
        return super().train(
            train_data_loader,
            val_data_loader,
            calib_data_loader,
            fit_config,
            calib_config,
            **fit_kwargs,
        )

    def calibrate(
        self,
        calib_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        calib_config: CalibConfig = CalibConfig(),
    ) -> Status:
        """
        Calibrate the probabilistic classifier.

        Parameters
        ----------
        calib_data_loader : DataLoader
            A calibration data loader.
        val_data_loader : DataLoader
            A validation data loader.
        calib_config : CalibConfig
            An object to configure the calibration.

        Returns
        -------
        Status
            A calibration status object. It provides information about the calibration.
        """
        self._check_output_dim(calib_data_loader)
        if val_data_loader is not None:
            self._check_output_dim(val_data_loader)
        return super()._calibrate(
            uncertainty_fn=calib_config.monitor.uncertainty_fn
            if calib_config.monitor.uncertainty_fn is not None
            else self.prob_output_layer.mean,
            calib_data_loader=calib_data_loader,
            val_data_loader=val_data_loader,
            calib_config=calib_config,
        )



import abc
from typing import Optional

from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.typing import Array
from fortuna.utils.random import WithRNG


class ProbOutputLayer(WithRNG, abc.ABC):
    r"""
    Abstract probabilistic output layer class. It characterizes the distribution of the target variable given the
    calibrated outputs. It can be see as :math:`p(y|\omega)`, where :math:`y` is a target variable and :math:`\omega` a
    calibrated output. The probabilistic output layer is not join over different data points, and it acts on them
    individually.
    """

    @abc.abstractmethod
    def log_prob(self, outputs: Array, targets: Array, **kwargs) -> jnp.ndarray:
        """
        Evaluate the log-probability density function (a.k.a. log-pdf) of target variables for each of the outputs.

        Parameters
        ----------
        outputs : Array
            Calibrated outputs.
        targets : Array
            Target data points.

        Returns
        -------
        jnp.ndarray
            An evaluation of the log-pdf for each output.
        """
        pass

    @abc.abstractmethod
    def predict(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Predict target variables starting from the calibrated outputs.

        Parameters
        ----------
        outputs : Array
            Calibrated outputs.

        Returns
        -------
        jnp.ndarray
            A predictions for each output.
        """

    @abc.abstractmethod
    def sample(
        self,
        n_target_samples: int,
        outputs: Array,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> jnp.ndarray:
        """
        Sample target variables for each outputs.

        Parameters
        ----------
        n_target_samples: int
            The number of target samples to draw for each of the outputs.
        outputs : Array
            Calibrated outputs.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        jnp.ndarray
            Samples of the target variable for each output.
        """
        pass

    @abc.abstractmethod
    def mean(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the mean of the target variable given the output with respect to the probabilistic output layer
        distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated mean for each output.
        """
        pass

    @abc.abstractmethod
    def mode(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the mode of the target variable given the output with respect to the probabilistic output layer
        distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated mode for each output.
        """
        pass

    @abc.abstractmethod
    def variance(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the variance of the target variable given the output with respect to the probabilistic output layer
        distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated variance for each output.
        """
        pass

    def std(self, outputs: Array, variances: Optional[Array] = None) -> jnp.ndarray:
        """
        Estimate the standard deviation of the target variable given the output with respect to the probabilistic
        output layer distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs
        variances: Optional[Array]
            Variance for each output.

        Returns
        -------
        jnp.ndarray
            The estimated standard deviation for each output.
        """
        return jnp.sqrt(self.variance(outputs)) if variances is None else variances

    @abc.abstractmethod
    def entropy(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the entropy of the target variable given the output with respect to the probabilistic output layer
        distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated mean for each output.
        """
        pass



from typing import (
    Callable,
    Tuple,
    Union,
)

import jax.numpy as jnp
import numpy as np

from fortuna.conformal.regression.base import ConformalRegressor
from fortuna.typing import Array


class EnbPI(ConformalRegressor):
    def __init__(
        self, aggregation_fun: Callable[[Array], Array] = lambda x: jnp.mean(x, 0)
    ):
        """
        Ensemble Batch Prediction Intervals (EnbPI) is a conformal prediction algorithm for time series regression.
        By bootstrapping the data and training the model on each of the bootstrap samples, EnbPI is able to compute
        conformal intervals that satisfy an approximate marginal guarantee on each test data point. Furthermore,
        EnbPI can incorporate online feedback from incoming batches of data, and improve the conformal intervals
        without having to retrain the model.

        Parameters
        ----------
        aggregation_fun: Callable[[Array], Array]
            Aggregation function that takes a group of model predictions from different bootstrap samples and reduces
            them to a single one.
        """
        self.aggregation_fun = aggregation_fun

    def conformal_interval(
        self,
        bootstrap_indices: Array,
        bootstrap_train_preds: Array,
        bootstrap_test_preds: Array,
        train_targets: Array,
        error: float,
        return_residuals: bool = False,
    ) -> Union[Array, Tuple[Array, Array]]:
        """
        Compute a coverage interval for each of the test inputs of the time series, at the desired coverage error.
        This is supported only for one-dimensional target variables, and for one times series at the time.

        Parameters
        ----------
        bootstrap_indices: Array
            The indices, randomly sampled with replacement, of the training data in the time series used to train the
            model. The first dimension is over the different samples of indices. The second dimension contain the
            data points for each sample, which must be as many as thnumber of data points in the time series used for
            training. A simple way of obtaining indices randomly sampled with replacement is
            :code:`numpy.random.choice(T, size=(B, T))`,
            where :code:`T` is the number of training points in the time series,
            and :code:`B` is the number of bootstrap samples.
            It is the user job to make sure that the models are trained upon the data corresponding to the random
            indices.
        bootstrap_train_preds: Array
            Model predictions for each of the bootstrap samples of data of the time series used for training the model,
            evaluated at each of the training data inputs of the time series. The first dimension is over the
            different bootstrap samples. The second dimensions is over the training inputs.
            There may be a third dimension, corresponding to the dimensionality of the predictions,
            but if so this must be one.
        bootstrap_test_preds: Array
            Model predictions for each of the bootstrap samples of data of the time series used for training the model,
            evaluated at each of the test data inputs of the time series. The first dimension is over the
            different bootstrap samples. The second dimensions is over the test inputs. There may be a third dimension,
            corresponding to the dimensionality of the predictions, but if so this must be one.
        train_targets: Array
            The target variables of the training data points in the time series.
        error: float
            The desired coverage error. This must be a scalar between 0 and 1, extremes included.
        return_residuals: bool
            If True, return the residual errors computed over the training data. These are used in
            :meth:`~fortuna.conformal.regression.enbpi.EnbPI.conformal_interval_from_residuals`.

        Returns
        -------
        Union[Array, Tuple[Array, Array]]
            The conformal intervals. The two components of the second dimension correspond to the left and right
            interval bounds. If :code:`return_residuals` is set to True, then it returns also the residuals computed on
            the training set.
        """
        n_bootstraps, n_train_times = bootstrap_indices.shape
        if (
            jnp.min(bootstrap_indices) < 0
            or jnp.max(bootstrap_indices) >= n_train_times
        ):
            raise ValueError(
                f"All elements f `bootstrap_indices` must be integers from 0 to {n_train_times - 1} "
                f"corresponding to the indices of the data used for training in each of the bootstrap "
                f"samples."
            )
        if bootstrap_train_preds.shape[0] != n_bootstraps:
            raise ValueError(
                "The first dimension of `bootstrap_train_preds` and `bootstrap_indices` corresponds to "
                "the number of bootstrap samples, and must have the same size. However, "
                f"{bootstrap_train_preds.shape[0]} and {n_bootstraps} were found, respectively."
            )
        if bootstrap_train_preds.shape[1] != n_train_times:
            raise ValueError(
                "The second dimension of `bootstrap_train_preds` and `bootstrap_indices` corresponds to "
                "the number of data points in the time series used for training, "
                "and must have the same size. "
                f"However, {bootstrap_train_preds.shape[1]} and {n_train_times} were found, respectively."
            )
        if bootstrap_test_preds.shape[0] != n_bootstraps:
            raise ValueError(
                "The first dimension of `bootstrap_test_preds` and `bootstrap_indices` corresponds to the "
                "number of bootstrap samples, and must have the same size. However, "
                f"{bootstrap_test_preds.shape[0]} and {n_bootstraps} were found, respectively."
            )
        if bootstrap_train_preds.ndim == 3:
            if bootstrap_train_preds.shape[2] == 1:
                bootstrap_train_preds = bootstrap_train_preds.squeeze(2)
            else:
                raise ValueError(
                    "This method is supported only for scalar model predictions. However, `bootstrap_train_preds` has "
                    "third dimension greater than 1."
                )
        if bootstrap_test_preds.ndim == 3:
            if bootstrap_test_preds.shape[2] == 1:
                bootstrap_test_preds = bootstrap_test_preds.squeeze(2)
            else:
                raise ValueError(
                    "This method is supported only for scalar model predictions. However, `bootstrap_test_preds` has "
                    "third dimension greater than 1."
                )
        if train_targets.shape[0] != bootstrap_train_preds.shape[1]:
            raise ValueError(
                "The first dimension of `train_targets` and the second dimension of "
                "`bootstrap_train_preds` correspond to the number of data points in the time series used "
                "for training, and must have the same size. However, "
                f"{train_targets.shape[0]} and {bootstrap_train_preds.shape[1]} were found, respectively."
            )
        if train_targets.ndim == 2:
            if train_targets.shape[1] == 1:
                train_targets = train_targets.squeeze(1)
            else:
                raise ValueError(
                    "This method is supported only for scalar target variables. However, `train_targets` has "
                    "second dimension greater than 1."
                )

        in_bootstrap_indices = np.zeros((n_bootstraps, n_train_times), dtype=bool)
        np.put_along_axis(in_bootstrap_indices, bootstrap_indices, values=1, axis=1)
        aggr_bootstrap_test_preds = np.zeros(
            (n_train_times,) + bootstrap_test_preds.shape[1:]
        )
        train_residuals = np.zeros((n_train_times,) + train_targets.shape[1:])

        for t in range(n_train_times):
            which_bootstraps = np.where(~(in_bootstrap_indices[:, t]))[0]
            if len(which_bootstraps) > 0:
                aggr_bootstrap_train_pred = self.aggregation_fun(
                    bootstrap_train_preds[which_bootstraps, t]
                )
                train_residuals[t] = np.abs(
                    train_targets[t] - aggr_bootstrap_train_pred
                )
                aggr_bootstrap_test_preds[t] = self.aggregation_fun(
                    bootstrap_test_preds[which_bootstraps]
                )
            else:
                train_residuals[t] = np.abs(train_targets[t])

        test_quantiles = jnp.quantile(aggr_bootstrap_test_preds, q=1 - error, axis=0)
        residuals_quantile = jnp.quantile(train_residuals, q=1 - error, axis=0)

        left = test_quantiles - residuals_quantile
        right = test_quantiles + residuals_quantile

        conformal_intervals = jnp.array(list(zip(left, right)))
        if not return_residuals:
            return conformal_intervals
        return conformal_intervals, train_residuals

    def conformal_interval_from_residuals(
        self,
        train_residuals: Array,
        bootstrap_new_train_preds: Array,
        bootstrap_new_test_preds: Array,
        new_train_targets: Array,
        error: float,
    ) -> Union[Array, Tuple[Array, Array]]:
        """
        Compute a coverage interval for each of the test inputs of the time series, at the desired coverage error.
        This is supported only for one-dimensional target variables, and for one times series at the time.
        This method assumes the residuals over the training sets have already been computed via
        :meth:`~fortuna.conformal.regression.enbpi.EnbPI.conformal_interval`
        using the flag :code:`return_residuals=True`.
        If so, this method takes in the predictions for each bootstrap samples over the new incoming training and test
        data points, and computes conformal intervals without having to retrain the model. Compared to predicting
        conformal intervals over all test data points at once with
        :meth:`~fortuna.conformal.regression.enbpi.EnbPI.conformal_interval`,
        this method exploits incoming data points as a form of online feedback,
        and improves the conformal intervals accordingly.

        Parameters
        ----------
        train_residuals: Array
            Residuals over the training data points computed via
            :meth:`~fortuna.conformal.regression.enbpi.EnbPI.conformal_interval` using the flag
            :code:`return_residuals=True`.
        bootstrap_new_train_preds: Array
            Model predictions for each of the bootstrap samples of data of the time series used for training the model,
            evaluated at the new incoming training data inputs of the time series. The first dimension is over the
            different bootstrap samples. The second dimensions is over the new training inputs.
            There may be a third dimension, corresponding to the dimensionality of the predictions,
            but if so this must be one. As an example, suppose that the bootstrap has been done over data points from
            1 to T, and the model was trained over each bootstrap sample. A prediction was then made on a batch of
            data points from T+1 to T+T1. After the data from T+1 to T+T1 is observed, this will be used as the new
            batch of training data.
        bootstrap_new_test_preds: Array
            Model predictions for each of the bootstrap samples of data of the time series used for training the model,
            evaluated at the new incoming test data inputs of the time series. The first dimension is over the
            different bootstrap samples. The second dimensions is over the new test inputs.
            There may be a third dimension, corresponding to the dimensionality of the predictions,
            but if so this must be one. As an example, suppose that the bootstrap has been done over data points from
            1 to T, and the model was trained over each bootstrap sample. A prediction was then made on a batch of
            data points from T+1 to T+T1. After the data from T+1 to T+T1 is observed, data from T+T1+1 to T+2*T1 may
            be taken as the new batch of test data.
        new_train_targets: Array
            The target variables of the new batch of training data points in the time series.
        error: float
            The desired coverage error. This must be a scalar between 0 and 1, extremes included.
        Returns
        -------
        Union[Array, Tuple[Array, Array]]
            - The conformal intervals. The two components of the second dimension correspond to the left and right
            interval bounds.
            - A new set of residuals, which includes the residuals computed on the new batch of training data points.
            The number of stored training residuals is kept constant by eliminating the oldest ones.
        """
        if bootstrap_new_train_preds.shape[0] != bootstrap_new_test_preds.shape[0]:
            raise ValueError(
                "The first dimensions of `bootstrap_new_train_preds` and `bootstrap_new_test_preds` "
                "correspond to the number of bootstrap samples, "
                "and must have the same size. However, "
                f"{bootstrap_new_train_preds.shape[0]} and "
                f"{bootstrap_new_test_preds.shape[0]} were found, respectively."
            )
        if bootstrap_new_train_preds.ndim == 3:
            if bootstrap_new_train_preds.shape[2] == 1:
                bootstrap_new_train_preds = bootstrap_new_train_preds.squeeze(2)
            else:
                raise ValueError(
                    "This method is supported only for scalar model predictions. However, "
                    "`bootstrap_new_train_preds` has third dimension greater than 1."
                )
        if bootstrap_new_test_preds.ndim == 3:
            if bootstrap_new_test_preds.shape[2] == 1:
                bootstrap_new_test_preds = bootstrap_new_test_preds.squeeze(2)
            else:
                raise ValueError(
                    "This method is supported only for scalar model predictions. "
                    "However, `bootstrap_new_test_preds` has third dimension greater than 1."
                )
        if new_train_targets.shape[0] != bootstrap_new_train_preds.shape[1]:
            raise ValueError(
                "The first dimension of `new_train_targets` and the second dimension of "
                "`bootstrap_new_train_preds` correspond to the number of data points in the time series "
                "used for training, and must have the same size. However, "
                f"{new_train_targets.shape[0]} and {bootstrap_new_train_preds.shape[1]} were found, "
                f"respectively."
            )
        if new_train_targets.ndim == 2:
            if new_train_targets.shape[1] == 1:
                new_train_targets = new_train_targets.squeeze(1)
            else:
                raise ValueError(
                    "This method is supported only for scalar target variables. However, `new_train_targets` has "
                    "second dimension greater than 1."
                )

        train_residuals[:-1] = train_residuals[1:]
        train_residuals[-1] = np.abs(
            new_train_targets - self.aggregation_fun(bootstrap_new_train_preds)
        )
        aggr_test_preds = self.aggregation_fun(bootstrap_new_test_preds)
        residuals_quantile = jnp.quantile(train_residuals, q=1 - error, axis=0)

        left = aggr_test_preds - residuals_quantile
        right = aggr_test_preds + residuals_quantile

        return jnp.array(list(zip(left, right))), train_residuals



import abc
from typing import (
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.prob_output_layer.base import ProbOutputLayer
from fortuna.typing import (
    Array,
    CalibMutable,
    CalibParams,
)
from fortuna.utils.random import WithRNG


class Predictive(WithRNG, abc.ABC):
    def __init__(
        self,
        output_calib_manager: OutputCalibManager,
        prob_output_layer: ProbOutputLayer,
    ):
        r"""
        Abstract predictive distribution. It characterizes the distribution of the target variable given the
        calibrated outputs. It can be see as :math:`p(y|\omega)`, where :math:`y` is a target variable and
        :math:`\omega` a calibrated output.
        """
        self.output_calib_manager = output_calib_manager
        self.prob_output_layer = prob_output_layer
        self.state = None

    def log_prob(
        self, outputs: Array, targets: Array, calibrated: bool = True, **kwargs
    ) -> jnp.ndarray:
        """
        Evaluate the log-probability density function (a.k.a. log-pdf) given the outputs and target data.

        Parameters
        ----------
        outputs : Array
            Calibrated outputs.
        targets : Array
            Target data points.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            An evaluation of the log-pdf for each data point.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.log_prob(outputs, targets, **kwargs)

    def sample(
        self,
        n_samples: int,
        outputs: Array,
        rng: Optional[PRNGKeyArray] = None,
        calibrated: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        """
        Sample target variables for each outputs.

        Parameters
        ----------
        n_samples: int
            The number of target samples to draw for each of the outputs.
        outputs : Array
            Calibrated outputs.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            Samples of the target variable for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.sample(n_samples, outputs, rng, **kwargs)

    def mean(self, outputs: Array, calibrated: bool = True, **kwargs) -> jnp.ndarray:
        """
        Estimate the mean of the target variable given the output, with respect to the predictive distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated mean for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.mean(outputs, **kwargs)

    def mode(self, outputs: Array, calibrated: bool = True, **kwargs) -> jnp.ndarray:
        """
        Estimate the mode of the target variable given the output, with respect to the predictive distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated mode for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.mode(outputs, **kwargs)

    def variance(
        self, outputs: jnp.ndarray, calibrated: bool = True, **kwargs
    ) -> jnp.ndarray:
        """
        Estimate the variance of the target variable given the output, with respect to the predictive distribution.

        Parameters
        ----------
        outputs : jnp.ndarray
            Model outputs.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated variance for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.variance(outputs, **kwargs)

    def std(
        self,
        outputs: jnp.ndarray,
        variances: Optional[jnp.ndarray] = None,
        calibrated: bool = True,
    ) -> jnp.ndarray:
        """
        Estimate the standard deviation of the target variable given the output, with respect to the predictive
        distribution.

        Parameters
        ----------
        outputs : jnp.ndarray
            Model outputs.
        variances: Optional[jnp.ndarray]
            Variance for each output.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated standard deviation for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.std(outputs, variances=variances)

    def entropy(
        self, outputs: jnp.ndarray, calibrated: bool = True, **kwargs
    ) -> jnp.ndarray:
        """
        Estimate the entropy of the target variable given the output, with respect to the predictive distribution.

        Parameters
        ----------
        outputs : jnp.ndarray
            Model outputs.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated mean for each output.
        """
        if calibrated:
            self._check_calibrated()
            state = self.state.get()
            outputs = self.output_calib_manager.apply(
                params=state.params["output_calibrator"],
                outputs=outputs,
                mutable=state.mutable["output_calibrator"],
            )
        return self.prob_output_layer.entropy(outputs, **kwargs)

    def _check_calibrated(self) -> None:
        """
        Check that the model has been calibrated beforehand.
        """
        if self.state is None:
            raise ValueError(
                "No calibration state was found. The model must be calibrated beforehand."
            )

    def _log_joint_prob(
        self,
        params: CalibParams,
        targets: Array,
        outputs: Array,
        mutable: Optional[CalibMutable] = None,
        rng: Optional[PRNGKeyArray] = None,
        return_aux: Optional[List[str]] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Any]]:
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs", "mutable"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise AttributeError(
                """The auxiliary objects {} is unknown. Please make sure that all elements of `return_aux`
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )
        aux = dict()
        outs = self.output_calib_manager.apply(
            params=params["output_calibrator"],
            outputs=outputs,
            mutable=mutable["output_calibrator"],
            rng=rng,
            calib="mutable" in return_aux,
        )
        if (
            mutable is not None
            and mutable["output_calibrator"] is not None
            and "mutable" in return_aux
        ):
            outputs, aux["mutable"] = outs
            aux["mutable"] = dict(output_calibrator=aux["mutable"])
        else:
            outputs = outs
            if "mutable" in return_aux:
                aux["mutable"] = dict(output_calibrator=None)
        log_joint_prob = self.prob_output_layer.log_prob(outputs, targets).sum()

        if len(return_aux) == 0:
            return log_joint_prob
        else:
            if "outputs" in return_aux:
                aux["outputs"] = outputs
            return log_joint_prob, aux



from typing import Optional

from fortuna.prob_model.posterior.name_to_posterior_state import NameToPosteriorState
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.training.mixin import WithCheckpointingMixin
from fortuna.typing import (
    OptaxOptimizer,
    Path,
)


class WithPosteriorCheckpointingMixin(WithCheckpointingMixin):
    def restore_checkpoint(
        self,
        restore_checkpoint_path: Path,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        name_to_train_state: NameToPosteriorState = NameToPosteriorState,
        **kwargs,
    ) -> PosteriorState:
        return super().restore_checkpoint(
            restore_checkpoint_path,
            optimizer,
            prefix,
            name_to_train_state=name_to_train_state,
        )



import numpy as np
from scipy.optimize import brute

from fortuna.calibration.binary_classification.platt_scaling.base import (
    BaseBinaryClassificationPlattScaling,
)


class F1BinaryClassificationPlattScaling(
    BaseBinaryClassificationPlattScaling
):
    """
    A Platt scaling class for binary classification.
    It learns a linear transformation of the probability that the target variable is positive.
    The method attempts to maximize the F1 score.
    """

    def __init__(self):
        super().__init__()
        self._threshold = None

    def fit(self, probs: np.ndarray, targets: np.ndarray, threshold: float):
        self._check_probs(probs)
        self._check_targets(targets)

        self._threshold = threshold
        n_pos_targets = np.sum(targets)

        def loss_fn(params):
            temp_preds = np.clip(params[0] + params[1] * probs, 0, 1) >= threshold
            n_pos_preds = np.sum(temp_preds)
            n_joint = np.sum(targets * temp_preds)
            prec = n_joint / n_pos_preds if n_pos_preds > 0 else 0.0
            rec = n_joint / n_pos_targets
            if prec + rec == 0.0:
                return 0.0
            return -2 * prec * rec / (prec + rec)

        self._intercept, self._slope = brute(
            loss_fn, ranges=[(-1, 1), (np.min(probs), 1 / threshold)], Ns=100
        )

    def predict(self, probs: np.ndarray):
        self._check_probs(probs)
        return (self.predict_proba(probs) >= self._threshold).astype(int)

    @property
    def threshold(self):
        return self._threshold



import numpy as np
from scipy.optimize import brute

from fortuna.calibration.binary_classification.temp_scaling.base import (
    BaseBinaryClassificationTemperatureScaling,
)


class CrossEntropyBinaryClassificationTemperatureScaling(
    BaseBinaryClassificationTemperatureScaling
):
    """
    A temperature scaling class for binary classification.
    It scales the probability that the target variables is positive with a single learnable parameters.
    The method minimizes the binary cross-entropy loss.
    """

    def fit(self, probs: np.ndarray, targets: np.ndarray):
        self._check_probs(probs)
        self._check_targets(targets)

        def temp_scaling_fn(tau):
            temp_probs = np.clip(probs / tau, 1e-9, 1 - 1e-9)
            return -np.mean(
                targets * np.log(temp_probs) + (1 - targets) * np.log(1 - temp_probs)
            )

        self._temperature = brute(
            temp_scaling_fn, ranges=[(np.min(probs), 10)], Ns=1000
        )[0]



from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
)

from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.likelihood.base import Likelihood
from fortuna.typing import (
    Batch,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
    Targets,
)
from fortuna.utils.random import WithRNG


class Loss(WithRNG):
    def __init__(
        self,
        likelihood: Likelihood,
        loss_fn: Callable[[Callable, Targets], jnp.ndarray],
    ):
        self.likelihood = likelihood
        self.loss_fn = loss_fn

    def __call__(
        self,
        params: Params,
        batch: Batch,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        return_aux: Optional[List[str]] = None,
        train: bool = False,
        outputs: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Tuple[jnp.ndarray, Any]:
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs", "mutable", "calib_mutable"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise AttributeError(
                """The auxiliary objects {} is unknown. Please make sure that all elements of `return_aux`
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )
        if train and outputs is not None:
            raise ValueError(
                """When `outputs` is available, `train` must be set to `False`."""
            )
        if "mutable" in return_aux and outputs is not None:
            raise ValueError(
                """When `outputs` is available, `return_aux` cannot contain 'mutable'`."""
            )
        if not train and "mutable" in return_aux:
            raise ValueError(
                "Returning an auxiliary mutable is supported only during training. Please either set `train` to "
                "`True`, or remove 'mutable' from `return_aux`."
            )
        if "mutable" in return_aux and mutable is None:
            raise ValueError(
                "In order to be able to return an auxiliary mutable, an initial mutable must be passed as `mutable`. "
                "Please either remove 'mutable' from `return_aux`, or pass an initial mutable as `mutable`."
            )
        if "mutable" not in return_aux and mutable is not None and train is True:
            raise ValueError(
                """You need to add `mutable` to `return_aux`. When you provide a (not null) `mutable` variable during
                training, that variable will be updated during the forward pass."""
            )

        if outputs is None:
            outs = self.likelihood.model_manager.apply(
                params,
                batch[0],
                train=train,
                mutable=mutable,
                rng=rng,
            )
            if "mutable" in return_aux:
                outputs, aux = outs
                mutable = aux["mutable"]
            else:
                outputs = outs

        aux = dict()
        if self.likelihood.output_calib_manager is not None:
            outs = self.likelihood.output_calib_manager.apply(
                params=calib_params["output_calibrator"]
                if calib_params is not None
                else None,
                mutable=calib_mutable["output_calibrator"]
                if calib_mutable is not None
                else None,
                outputs=outputs,
                calib="calib_mutable" in return_aux,
            )
        if (
            calib_mutable is not None
            and calib_mutable["output_calibrator"] is not None
            and "calib_mutable" in return_aux
        ):
            outputs, aux["calib_mutable"] = outs
            aux["calib_mutable"] = dict(output_calibrator=aux["calib_mutable"])
        else:
            outputs = outs
            if "calib_mutable" in return_aux:
                aux["calib_mutable"] = dict(output_calibrator=None)

        if "outputs" in return_aux:
            aux["outputs"] = outputs
        if "mutable" in return_aux:
            aux["mutable"] = mutable

        return self.loss_fn(outputs, batch[1]), aux



from typing import NamedTuple

import jax
from jax._src.prng import PRNGKeyArray
import optax
from optax import GradientTransformation

from fortuna.prob_model.posterior.sgmcmc.sghmc.sghmc_integrator import sghmc_integrator
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_preconditioner import Preconditioner
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_step_schedule import (
    cyclical_cosine_schedule_with_const_burnin,
)


class OptaxCyclicalSGLDState(NamedTuple):
    """Optax state for the Cyclical SGLD integrator."""

    sgd_state: NamedTuple
    sgld_state: NamedTuple


def cyclical_sgld_integrator(
    rng_key: PRNGKeyArray,
    init_step_size: float,
    cycle_length: int,
    exploration_ratio: float,
    preconditioner: Preconditioner,
) -> GradientTransformation:
    """Optax implementation of the Cyclical SGLD integrator.

    Parameters
    ----------
        rng_key: PRNGKeyArray
            An initial random number generator.
        init_step_size: float
            The initial step size.
        cycle_length: int
            The length of each exploration/sampling cycle, in steps.
        exploration_ratio: float
            The fraction of steps to allocate to the mode exploration phase.
        preconditioner: Preconditioner
            See :class:`Preconditioner` for reference.
    """
    step_schedule = cyclical_cosine_schedule_with_const_burnin(
        init_step_size=init_step_size,
        burnin_steps=0,
        cycle_length=cycle_length,
    )
    # SGHMC with no momentum is equivalent to SGLD
    sgld = sghmc_integrator(
        momentum_decay=0.0,
        momentum_resample_steps=None,
        rng_key=rng_key,
        step_schedule=step_schedule,
        preconditioner=preconditioner,
    )
    sgd = optax.sgd(learning_rate=1.0)

    def init_fn(params):
        return OptaxCyclicalSGLDState(
            sgd_state=sgd.init(params),
            sgld_state=sgld.init(params),
        )

    def update_fn(gradient, state, *_):
        def sgd_step():
            step_size = step_schedule(state.sgld_state.count)
            preconditioner_state = preconditioner.update_preconditioner(
                gradient, state.sgld_state.preconditioner_state
            )
            new_sgld_state = state.sgld_state._replace(
                count=state.sgld_state.count + 1,
                preconditioner_state=preconditioner_state,
            )
            rescaled_gradient = jax.tree_map(
                lambda g: -1.0 * step_size * g,
                gradient,
            )
            updates, new_sgd_state = sgd.update(rescaled_gradient, state.sgd_state)
            updates = preconditioner.multiply_by_m_inv(updates, preconditioner_state)
            new_state = OptaxCyclicalSGLDState(
                sgd_state=new_sgd_state,
                sgld_state=new_sgld_state,
            )
            return updates, new_state

        def sgld_step():
            updates, new_sgld_state = sgld.update(gradient, state.sgld_state)
            new_state = OptaxCyclicalSGLDState(
                sgd_state=state.sgd_state,
                sgld_state=new_sgld_state,
            )
            return updates, new_state

        updates, state = jax.lax.cond(
            ((state.sgld_state.count % cycle_length) / cycle_length)
            >= exploration_ratio,
            sgld_step,
            sgd_step,
        )
        return updates, state

    return GradientTransformation(init_fn, update_fn)



from __future__ import annotations

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

import jax.numpy as jnp

from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import (
    AnyKey,
    Array,
    OptaxOptimizer,
)
from fortuna.utils.strings import (
    convert_string_to_tuple,
    encode_tuple_of_lists_of_strings_to_numpy,
)


class CyclicalSGLDState(PosteriorState):
    """
    Attributes
    ----------
    encoded_name: jnp.ndarray
        CyclicalSGLDState state name encoded as an array.
    """

    encoded_name: tuple = convert_string_to_tuple("CyclicalSGLDState")
    _encoded_which_params: Optional[Dict[str, List[Array]]] = None

    @classmethod
    def convert_from_map_state(
        cls,
        map_state: MAPState,
        optimizer: OptaxOptimizer,
        which_params: Tuple[List[AnyKey], ...],
    ) -> CyclicalSGLDState:
        """
        Convert a MAP state into an CyclicalSGLDState state.

        Parameters
        ----------
        map_state: MAPState
            A MAP posterior state.
        optimizer: OptaxOptimizer
            An Optax optimizer.
        which_params: Tuple[List[AnyKey], ...]
            Sequences of keys pointing to the stochastic parameters.

        Returns
        -------
        CyclicalSGLDState
            An Cyclical SGLD state.
        """
        _encoded_which_params = encode_tuple_of_lists_of_strings_to_numpy(which_params)
        return cls.init(
            params=map_state.params,
            mutable=map_state.mutable,
            optimizer=optimizer,
            calib_params=map_state.calib_params,
            calib_mutable=map_state.calib_mutable,
            _encoded_which_params=_encoded_which_params,
        )



from typing import (
    List,
    Tuple,
)

from flax.core import FrozenDict
import jax
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree
from jax.tree_util import tree_map

from fortuna.data.loader import DataLoader
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_state import ADVIState
from fortuna.prob_model.posterior.normalizing_flow.normalizing_flow_state import (
    NormalizingFlowState,
)
from fortuna.prob_model.posterior.normalizing_flow.normalizing_flow_trainer import (
    NormalizingFlowTrainer,
)
from fortuna.training.trainer import (
    JittedMixin,
    MultiDeviceMixin,
)
from fortuna.typing import (
    Params,
    Path,
)
from fortuna.utils.freeze import (
    get_frozen_paths,
    get_trainable_paths,
)
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_pair,
    nested_set,
)


class ADVITrainer(NormalizingFlowTrainer):
    def __str__(self):
        return ADVI_NAME

    def save_checkpoint(
        self,
        state: ADVIState,
        save_checkpoint_dir: Path,
        keep: int = 1,
        force_save: bool = False,
        prefix: str = "checkpoint_",
    ) -> None:
        state = state.replace(
            params=self._unravel_params(state.params),
            frozen_params=FrozenDict(),
            _encoded_which_params=self._encoded_which_params,
        )
        super().save_checkpoint(state, save_checkpoint_dir, keep, force_save, prefix)

    def on_train_end(self, state: NormalizingFlowState) -> NormalizingFlowState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )

        state = state.replace(
            params=self._unravel_params(state.params),
            frozen_params=None,
            _encoded_which_params=self._encoded_which_params,
        )
        return state

    def _unravel_params(
        self,
        rav_params: FrozenDict,
    ) -> Params:
        if self._which_params is not None:
            return FrozenDict(
                nested_pair(
                    self._unravel(rav_params["mean"]).unfreeze(),
                    self._which_params,
                    self._sub_unravel(rav_params["log_std"]),
                    ("mean", "log_std"),
                )
            )

        params = self._unravel(rav_params["mean"]).unfreeze()
        log_stds = self._unravel(rav_params["log_std"])
        for k, v in params.items():
            params[k] = {
                "params": dict(mean=v["params"], log_std=log_stds[k]["params"])
            }
        return FrozenDict(params)

    def on_train_start(
        self,
        state: NormalizingFlowState,
        dataloaders: List[DataLoader],
        rng: PRNGKeyArray,
    ) -> Tuple[NormalizingFlowState, List[DataLoader], PRNGKeyArray]:
        if self.freeze_fun is not None:
            frozen_paths = get_frozen_paths(
                self._unravel(state.params["mean"]), self.freeze_fun
            )
            trainable_paths = get_trainable_paths(
                self._unravel(state.params["mean"]), self.freeze_fun
            )

            state = state.replace(
                frozen_params=FrozenDict(
                    {
                        s: ravel_pytree(
                            nested_set(
                                d={},
                                key_paths=frozen_paths,
                                objs=tuple(
                                    [
                                        nested_get(
                                            self._unravel(state.params[s]).unfreeze(),
                                            path,
                                        )
                                        for path in frozen_paths
                                    ]
                                ),
                                allow_nonexistent=True,
                            )
                        )[0]
                        for s in ["mean", "log_std"]
                    }
                ),
                params=FrozenDict(
                    {
                        s: ravel_pytree(
                            nested_set(
                                d={},
                                key_paths=trainable_paths,
                                objs=tuple(
                                    [
                                        nested_get(
                                            self._unravel(state.params[s]).unfreeze(),
                                            path,
                                        )
                                        for path in trainable_paths
                                    ]
                                ),
                                allow_nonexistent=True,
                            )
                        )[0]
                        for s in ["mean", "log_std"]
                    }
                ),
            )

        return state, dataloaders, rng


class JittedADVITrainer(JittedMixin, ADVITrainer):
    pass


class MultiDeviceADVITrainer(MultiDeviceMixin, ADVITrainer):
    def on_train_end(self, state: ADVIState) -> ADVIState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )

        state = jax.device_get(tree_map(lambda x: x[0], state))
        state = state.replace(
            params=self._unravel_params(state.params),
            frozen_params=None,
            _encoded_which_params=self._encoded_which_params,
        )
        return state



from fortuna.hallucination.base import HallucinationMulticalibrator
from fortuna.hallucination.grouping.clustering.base import GroupingModel
from fortuna.hallucination.scoring.inv_perplexity import inv_perplexity



from typing import (
    Callable,
    Dict,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
from jax import (
    jacrev,
    vmap,
)
import jax.numpy as jnp
from jax.tree_util import (
    tree_map,
    tree_reduce,
)

from fortuna.typing import (
    AnyKey,
    Array,
    InputData,
    Params,
)
from fortuna.utils.freeze import get_paths_with_label
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)


def value_and_jacobian_squared_row_norm(
    apply_fn: Callable[
        [Params, InputData], Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]
    ],
    params: Params,
    x: InputData,
    has_aux: bool = False,
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]] = None,
    top_k: Optional[int] = None,
) -> Tuple[Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]], jnp.ndarray]:
    params = params.unfreeze()

    params_paths = None
    sub_params = None
    if freeze_fun is not None:
        params_paths = tuple(
            get_paths_with_label(
                params, freeze_fun, label=True, allowed_labels=[True, False]
            )
        )
        sub_params = tuple([nested_get(d=params, keys=path) for path in params_paths])

    def set_params(_p):
        if params_paths is None:
            return _p
        return FrozenDict(nested_set(d=params, key_paths=params_paths, objs=_p))

    def _apply_fn(_p, _x):
        _f = apply_fn(set_params(_p), _x)
        if has_aux:
            _f, _ = _f
        return _f[0]

    f = apply_fn(params, x)
    if has_aux:
        f, aux = f

    n_dim = f.shape[-1]

    indices = None
    if top_k is not None:
        indices = vmap(lambda _f: jnp.argsort(_f)[-top_k:])(f)

    x = x[:, None] if not isinstance(x, dict) else tree_map(lambda v: v[:, None], x)

    @vmap
    def jacobian_squared_row_norm_fn(_x, idx):
        _row_norms = tree_reduce(
            lambda a, b: a + jnp.sum(b**2, axis=tuple(range(1, b.ndim))),
            jacrev(
                lambda p: _apply_fn(p, _x)[idx] if idx is not None else _apply_fn(p, _x)
            )(sub_params if params_paths is not None else params),
            initializer=0,
        )
        if idx is None:
            return _row_norms
        row_norms = jnp.max(_row_norms) * jnp.ones(n_dim)
        row_norms = row_norms.at[idx].set(_row_norms)
        return row_norms

    jac_squared_row_norm = jacobian_squared_row_norm_fn(x, indices)

    if has_aux:
        return (f, aux), jac_squared_row_norm
    return f, jac_squared_row_norm



import abc
import logging
from typing import (
    Dict,
    Optional,
    Tuple,
    Union,
)

from jax import vmap
import jax.numpy as jnp

from fortuna.conformal.multivalid.base import MultivalidMethod
from fortuna.typing import Array


class OneShotMultivalidMethod(MultivalidMethod):
    def __init__(self, seed: int = 0):
        """
        A base one-shot multivalid method.

        Parameters
        ----------
        seed: int
            Random seed.
        """
        super().__init__(seed=seed)
        self._patches = None
        self._n_buckets = None

    def calibrate(
        self,
        scores: Array,
        values: Array = None,
        test_values: Optional[Array] = None,
        n_buckets: int = 100,
        min_prob_b: Union[float, str] = "auto",
    ) -> Union[None, Array]:
        """
        Calibrate the model by finding a list of patches to the model that bring the calibration error below a
        certain threshold.

        Parameters
        ----------
        scores: Array
            A list of scores :math:`s(x, y)` computed on the calibration data.
            This should be a one-dimensional array of elements between 0 and 1.
        values: Optional[Array]
            The initial model evalutions :math:`f(x)` on the calibration data. If not provided, these are set to 0.
        test_values: Optional[Array]
            The initial model evaluations :math:`f(x)` on the test data. If not provided, these are set to 0.
        n_buckets: int
            The number of buckets used in the algorithm.
        min_prob_b: Union[float, str]
            Minimum probability of the conditioning set :math:`B_t` for the patch to be applied.
            If "auto", it will be chosen based on the number of buckets and dimension of the scores.

        Returns
        -------
        Union[Dict, Tuple[Array, Dict]]
            A status including the number of rounds taken to reach convergence and the calibration errors computed
            during the training procedure. if `test_values` is provided, the list of patches will
            be applied, and the calibrated test values will be returned together with the status.
        """
        if test_values is not None and values is None:
            raise ValueError(
                "If `test_values is provided, `values` must also be provided."
            )
        if min_prob_b != "auto" and (min_prob_b < 0 or min_prob_b > 1):
            raise ValueError(
                "`min_prob_b` must be greater than or equal to 0 and less than or equal to 1."
            )
        if values is None:
            raise ValueError("`values` must be provided.")

        self._check_scores(scores)
        scores = self._process_scores(scores)
        n_dims = scores.shape[1]

        min_prob_b = self._maybe_init_min_prob_b(
            min_prob_b=min_prob_b, n_buckets=n_buckets, n_dims=n_dims
        )

        self._maybe_check_values(values, test_values)
        values, test_values = self._maybe_process_values(values, test_values)

        self.n_buckets = n_buckets
        buckets = self._get_buckets(n_buckets)

        self._patches = vmap(
            lambda v: vmap(
                lambda c: self._get_patch(
                    v=v, c=c, scores=scores[:, c], values=values, min_prob_b=min_prob_b
                )
            )(jnp.arange(n_dims))
        )(buckets)

        if test_values is not None:
            return self.apply_patches(test_values)

    def apply_patches(
        self,
        values: Array,
    ) -> Array:
        if self._patches is None:
            logging.warning("No patches available.")
            return values

        buckets = self._get_buckets(self.n_buckets)
        unique_values = jnp.unique(values)

        n_dims = 1 if values.ndim == 1 else values.shape[1]

        b = vmap(
            lambda v: vmap(
                lambda c: self._get_b(values, v, c, self.n_buckets), out_axes=1
            )(jnp.arange(n_dims)),
            out_axes=1,
        )(unique_values)

        return self._patch(
            values=values, b=b, unique_values=unique_values, buckets=buckets
        )

    @staticmethod
    def _get_b(
        values: Array,
        v: Array,
        c: Optional[Array],
        n_buckets: int,
    ) -> Array:
        return jnp.abs(values - v) < 0.5 / n_buckets

    def _patch(
        self, values: Array, b: Array, unique_values: Array, buckets: Array
    ) -> Array:
        patched_values = jnp.copy(values)
        for i, v in enumerate(unique_values):
            for c in range(b.shape[2]):
                idx_v = jnp.argmin(jnp.abs(buckets - v))
                if values.ndim == 1:
                    patched_values = patched_values.at[b[:, i, c]].set(
                        self._patches[idx_v, c]
                    )
                else:
                    patched_values = patched_values.at[b[:, i, c], c].set(
                        self._patches[idx_v, c]
                    )
        return patched_values

    @abc.abstractmethod
    def _get_patch(
        self, v: Array, c: Array, scores: Array, values: Array, min_prob_b: float
    ) -> Array:
        pass



from typing import (
    Dict,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from flax import linen as nn
from flax.core import FrozenDict
from flax.training.checkpoints import PyTree
import jax
from jax import (
    numpy as jnp,
    random,
)
from jax._src.prng import PRNGKeyArray

from fortuna.model.model_manager.classification import (
    ClassificationModelManager,
    SNGPClassificationModelManagerMixin,
)
from fortuna.model_editor.base import ModelEditor
from fortuna.typing import (
    Array,
    Mutable,
    Params,
)
from fortuna.utils.data import get_inputs_from_shape
from fortuna.utils.nested_dicts import nested_update


class HuggingFaceClassificationModelManager(ClassificationModelManager):
    def apply(
        self,
        params: Params,
        inputs: Dict[str, Array],
        mutable: Optional[Mutable] = None,
        train: bool = False,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        # setup dropout key
        if rng is not None:
            rng, dropout_rng = random.split(rng, 2)
        else:
            dropout_rng = None

        model_kwargs = {}
        if mutable is not None:
            model_kwargs = {"mutable": mutable}

        has_aux = train and mutable

        def apply_fn(p, x):
            _outputs = self.model(
                **x,
                params=p,
                dropout_rng=dropout_rng,
                train=train,
                output_attentions=kwargs.get("output_attentions"),
                output_hidden_states=kwargs.get("output_hidden_states"),
                return_dict=kwargs.get("return_dict"),
                **model_kwargs,
            )
            if hasattr(_outputs, "logits"):
                _outputs = _outputs.logits

            if isinstance(_outputs, tuple) and not has_aux:
                _outputs = _outputs[0]
            return _outputs

        if self.model_editor is not None:
            outputs = self.model_editor.apply(
                params["model_editor"],
                apply_fn=apply_fn,
                model_params=params["model"]["params"],
                x=inputs,
                has_aux=has_aux,
            )
        else:
            outputs = apply_fn(params["model"]["params"], inputs)

        if has_aux:
            outputs, mutable = outputs
            return outputs, {"mutable": FrozenDict({"model": mutable})}
        return outputs

    def init(
        self, input_shape: Tuple[int, ...], rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, Mapping]:
        assert self.model._is_initialized, (
            "At the moment Fortuna supports models from Hugging Face that are loaded via "
            "`from_pretrained` method, which also takes care of model initialization."
        )
        params = {"model": {"params": self.model.params}}
        if self.model_editor is not None:
            if rng is None:
                rng = self.rng.get()
            rng, params_key, dropout_key = random.split(rng, 3)
            rngs = {"params": params_key, "dropout": dropout_key}

            def apply_fn(p, x):
                _outputs = self.model(**x, params=p)
                if hasattr(_outputs, "logits"):
                    _outputs = _outputs.logits
                return _outputs

            params.update(
                dict(
                    model_editor=self.model_editor.init(
                        rngs,
                        apply_fn=apply_fn,
                        model_params=FrozenDict(params["model"]["params"]),
                        x=get_inputs_from_shape(input_shape),
                        has_aux=False,
                    )
                )
            )
        return params


class SNGPHuggingFaceClassificationModelManager(
    SNGPClassificationModelManagerMixin, HuggingFaceClassificationModelManager
):
    def __init__(
        self,
        model: nn.Module,
        model_editor: Optional[ModelEditor] = None,
        *args,
        **kwargs,
    ):
        super(SNGPHuggingFaceClassificationModelManager, self).__init__(
            model, model_editor=model_editor, *args, **kwargs
        )

    def init(
        self, input_shape: Tuple[int, ...], rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, FrozenDict]:
        if rng is None:
            rng = self.rng.get()
        assert self.model._is_initialized, (
            "At the moment Fortuna supports models from Hugging Face that are loaded via "
            "`from_pretrained` method, which also takes care of model initialization."
        )
        output_shape = jax.eval_shape(
            self.model, **get_inputs_from_shape(input_shape)
        ).shape
        if len(output_shape[1:]) > 1:  # drop batch size
            raise ValueError(
                f"The output shape for the given model is {output_shape}.\n"
                f"In order to use SNGP the output shape of the provide model has to be of shape"
                f"(batch_size, n_features)."
            )
        rng, params_key, dropout_key = random.split(rng, 3)
        rngs = {"params": params_key, "dropout": dropout_key}
        gp_params = self._gp_output_model.init(rngs, jnp.zeros(output_shape), **kwargs)
        params = dict(model=nested_update(self.model.params, gp_params.unfreeze()))

        if self.model_editor is not None:
            rng, params_key, dropout_key = random.split(rng, 3)
            rngs = {"params": params_key, "dropout": dropout_key}

            mutable = dict(
                model=FrozenDict(
                    {k: v for k, v in params["model"].items() if k != "params"}
                )
            )

            def apply_fn(p, x):
                _outputs = self.model(**x, params=p, mutable=mutable)
                return _outputs

            params.update(
                dict(
                    model_editor=self.model_editor.init(
                        rngs,
                        apply_fn=apply_fn,
                        model_params=FrozenDict(params["model"]["params"]),
                        x=get_inputs_from_shape(input_shape),
                        has_aux=False,
                    )
                )
            )
        return params



from __future__ import annotations

from typing import (
    Dict,
    Optional,
    Tuple,
    Union,
)

from jax import random
import jax.numpy as jnp

from fortuna.conformal.multivalid.iterative.multicalibrator import Multicalibrator
from fortuna.conformal.multivalid.mixins.classification.top_label_multicalibrator import (
    TopLabelMulticalibratorMixin,
)
from fortuna.typing import Array


class TopLabelMulticalibrator(TopLabelMulticalibratorMixin, Multicalibrator):
    def __init__(self, n_classes: int, seed: int = 0):
        """
        A multicalibration method that provides multivalid coverage guarantees. See Algorithm 15 in `Aaron Roth's notes
        <https://www.cis.upenn.edu/~aaroth/uncertainty-notes.pdf>`_.

        Parameters
        ----------
        n_classes: int
            Number of classes.
        seed: int
            Random seed.
        """
        super().__init__(n_classes, seed=seed)

    def calibrate(
        self,
        targets: Array,
        groups: Optional[Array] = None,
        probs: Optional[Array] = None,
        test_groups: Optional[Array] = None,
        test_probs: Optional[Array] = None,
        atol: float = 1e-4,
        rtol: float = 1e-6,
        min_prob_b: Union[float, str] = "auto",
        n_buckets: int = 100,
        n_rounds: int = 1000,
        eta: float = 0.1,
        split: float = 0.8,
        bucket_types: Tuple[str, ...] = ("<=", ">="),
        patch_type: str = "additive",
        **kwargs,
    ) -> Union[Dict, Tuple[Array, Dict]]:
        return super().calibrate(
            scores=self._get_scores(targets),
            groups=groups,
            values=probs,
            test_groups=test_groups,
            test_values=test_probs,
            atol=atol,
            rtol=rtol,
            min_prob_b=min_prob_b,
            n_buckets=n_buckets,
            n_rounds=n_rounds,
            eta=eta,
            split=split,
            bucket_types=bucket_types,
            patch_type=patch_type,
            **kwargs,
        )

    def apply_patches(
        self,
        groups: Optional[Array] = None,
        probs: Optional[Array] = None,
    ) -> Array:
        return super().apply_patches(groups=groups, values=probs)

    def calibration_error(
        self,
        targets: Array,
        groups: Optional[Array] = None,
        probs: Optional[Array] = None,
        n_buckets: int = 10,
        **kwargs,
    ) -> Array:
        return super().calibration_error(
            scores=self._get_scores(targets),
            groups=groups,
            values=probs,
            n_buckets=n_buckets,
            **kwargs,
        )

    def _get_b(
        self,
        groups: Array,
        values: Array,
        v: Array,
        g: Array,
        c: Array,
        tau: Array,
        n_buckets: int,
    ) -> Array:
        b = super()._get_b(
            groups=groups,
            values=values[:, c],
            v=v,
            g=g,
            c=c,
            tau=tau,
            n_buckets=n_buckets,
        )
        b *= values.argmax(1) == c
        return b

    def _maybe_init_values(
        self, values: Optional[Array], size: Optional[int] = None
    ) -> Array:
        if values is None:
            if size is None:
                raise ValueError(
                    "If `values` is not provided, `size` must be provided."
                )
            values = 1 / self.n_classes * jnp.ones((size, self.n_classes))
            values += 0.01 * random.normal(
                random.PRNGKey(self._seed), shape=values.shape
            )
            values = jnp.abs(values)
            values -= 2 * jnp.maximum(0, values - 1)
            values /= values.sum(1, keepdims=True)
        else:
            values = jnp.copy(values)
        return values



import abc

import numpy as np


class BaseBinaryClassificationTemperatureScaling(abc.ABC):
    def __init__(self):
        self._temperature = None

    @abc.abstractmethod
    def fit(self, probs: np.ndarray, targets: np.ndarray):
        """
        Fit the temperature scaling method.

        Parameters
        ----------
        probs: np.ndarray
            A one-dimensional probabilities of positive target variables for each input.
        targets: np.ndarray
            A one-dimensional array of integer target variables for each input.
        """
        pass

    def predict_proba(self, probs: np.ndarray) -> np.ndarray:
        """
        Predict the scaled probabilities for each input.

        Parameters
        ----------
        probs: np.ndarray
            A one-dimensional probabilities of positive target variables for each input.
        Returns
        -------
        np.ndarray
            The predicted probabilities
        """
        self._check_probs(probs)
        return np.clip(probs / self._temperature, 0.0, 1.0)

    def predict(self, probs: np.ndarray, threshold: float = 0.5) -> np.ndarray:
        """
        Predict the target variable for each input.

        Parameters
        ----------
        probs: np.ndarray
            A one-dimensional probabilities of positive target variables for each input.
        threshold: np.ndarray
            The threshold on the predicted probabilities do decide whether a target variable is positive or
            negative.

        Returns
        -------
        np.ndarray
            The predicted target variables.
        """
        self._check_probs(probs)
        return (self.predict_proba(probs) >= threshold).astype(int)

    @property
    def temperature(self):
        return self._temperature

    @staticmethod
    def _check_probs(probs: np.ndarray):
        if probs.ndim != 1:
            raise ValueError("The array of probabilities must be one-dimensional.")

    @staticmethod
    def _check_targets(targets: np.ndarray):
        if targets.ndim != 1:
            raise ValueError("The array of targets must be one-dimensional.")
        if targets.dtype != int:
            raise ValueError("Each element in the array of targets must be an integer.")



from typing import (
    Callable,
    Optional,
)

from flax import linen as nn
import jax.numpy as jnp

from fortuna.calib_model.base import CalibModel
from fortuna.calib_model.config.base import Config
from fortuna.calib_model.predictive.regression import RegressionPredictive
from fortuna.data import DataLoader
from fortuna.likelihood.regression import RegressionLikelihood
from fortuna.loss.regression.scaled_mse import scaled_mse_fn
from fortuna.model.model_manager.regression import RegressionModelManager
from fortuna.model_editor.base import ModelEditor
from fortuna.prob_output_layer.regression import RegressionProbOutputLayer
from fortuna.typing import (
    Outputs,
    Status,
    Targets,
)


class CalibRegressor(CalibModel):
    def __init__(
        self,
        model: nn.Module,
        likelihood_log_variance_model: nn.Module,
        model_editor: Optional[ModelEditor] = None,
        seed: int = 0,
    ):
        r"""
        A calibration regressor class.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. It characterizes the mean model
            of the likelihood function. The outputs must belong to the same space as the target variables.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`\mu(w, x)`.
        likelihood_log_variance_model: nn.Module
            A model characterizing the log-variance of a Gaussian likelihood function. The outputs must belong to the
            same space as the target variables. Let :math:`x` be input variables and :math:`w` the random model
            parameters. Then the model is described by a function :math:`\log\sigma^2(w, x)`.
        model_editor : ModelEditor
            A model_editor objects. It takes the forward pass and transforms the outputs.
        seed: int
            A random seed.

        Attributes
        ----------
        model : nn.Module
            See `model` in `Parameters`.
        model_manager : RegressionModelManager
            This object orchestrates the model's forward pass. Given a mean model :math:`\mu(w, x)` and a log-variance
            model :math:`\log\sigma^2`, the model manager concatenates the two into
            :math:`f(w, x)=[\mu(w, x), \log\sigma^2(w, x)]`.
        prob_output_layer : RegressionProbOutputLayer
            This object characterizes the distribution of the target variable given the calibrated outputs. It is
            defined by :math:`p(y|\omega)=\text{Categorical}(p=softmax(\omega))`, where :math:`\omega` denote the
            calibrated outputs and :math:`y` denotes a target variable.
        likelihood : RegressionLikelihood
            The likelihood function. This is defined by
            :math:`p(y|w, \phi, x) = \text{Categorical}(p=\text{softmax}(g(\phi, f(w, x)))`.
        predictive : RegressionPredictive
            This denotes the predictive distribution, that is :math:`p(y|\phi, x, \mathcal{D})`.
        """
        self.model_manager = RegressionModelManager(
            model, likelihood_log_variance_model, model_editor=model_editor
        )
        self.prob_output_layer = RegressionProbOutputLayer()
        self.likelihood = RegressionLikelihood(
            model_manager=self.model_manager,
            prob_output_layer=self.prob_output_layer,
            output_calib_manager=None,
        )
        self.predictive = RegressionPredictive(likelihood=self.likelihood)
        super().__init__(seed=seed)

    def _check_output_dim(self, data_loader: DataLoader):
        data_output_dim = 0
        for x, y in data_loader:
            input_shape = x.shape[1:]
            data_output_dim = y.shape[1]
            break
        if data_output_dim == 0:
            raise ValueError(
                """`_data_loader` is either empty or incorrectly constructed."""
            )
        model_manager_output_dim = self._get_output_dim(input_shape)
        if model_manager_output_dim != 2 * data_output_dim:
            raise ValueError(
                f"""The outputs dimension of both `model` and `likelihood_log_variance_model` must be the same as
                the dimension of the target variables in `_data_loader`. However, {model_manager_output_dim // 2} and
                {data_output_dim} were found, respectively."""
            )

    def calibrate(
        self,
        calib_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray] = scaled_mse_fn,
        config: Config = Config(),
    ) -> Status:
        """
        Calibrate the calibration regressor.

        Parameters
        ----------
        calib_data_loader : DataLoader
            A calibration data loader.
        val_data_loader : DataLoader
            A validation data loader.
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray]
            The loss function to use for calibration.
        config : Config
            An object to configure the calibration.

        Returns
        -------
        Status
            A calibration status object. It provides information about the calibration.
        """
        self._check_output_dim(calib_data_loader)
        if val_data_loader is not None:
            self._check_output_dim(val_data_loader)
        return self._calibrate(
            calib_data_loader=calib_data_loader,
            uncertainty_fn=config.monitor.uncertainty_fn
            if config.monitor.uncertainty_fn is not None
            else self.prob_output_layer.variance,
            val_data_loader=val_data_loader,
            loss_fn=loss_fn,
            config=config,
        )
